import logging
from uuid import uuid4

from django.http import Http404

from rest_framework.generics import (
    RetrieveAPIView,
    DestroyAPIView,
    GenericAPIView
)

from rest_framework import status
from rest_framework.response import Response
from rest_framework_json_api.renderers import JSONRenderer
from rest_framework.parsers import JSONParser

from drf_spectacular.utils import extend_schema

from papermerge.core.models import Page, Document, Folder
from papermerge.core.lib.utils import (
    get_reordered_list,
    annotate_page_data
)
from papermerge.core.utils import clock

from papermerge.core.storage import get_storage_instance
from papermerge.core.serializers import (
    PageSerializer,
    PageDeleteSerializer,
    PagesReorderSerializer,
    PagesRotateSerializer,
    PagesMoveToDocumentSerializer,
    PagesMoveToFolderSerializer
)
from papermerge.core.renderers import (
    PlainTextRenderer,
    ImageJpegRenderer,
    ImageSVGRenderer
)
from papermerge.core.exceptions import APIBadRequest
from papermerge.core.signal_definitions import (
    page_move_to_folder,
    page_move_to_document,
    page_rotate,
    page_delete,
    page_reorder
)
from .mixins import RequireAuthMixin
from .utils import (
    remove_pdf_pages,
    insert_pdf_pages,
    total_merge,
    partial_merge,
    reuse_ocr_data,
    reuse_ocr_data_multi,
    reuse_text_field,
    reuse_text_field_multi,
    reorder_pdf_pages,
    rotate_pdf_pages,
    PageRecycleMap
)
from ..models.utils import OCR_STATUS_SUCCEEDED

logger = logging.getLogger(__name__)


def reuse_ocr_data_after_rotate(
    old_version,
    new_version,
    pages_data,
    pages_count
):
    pass


class PageView(RequireAuthMixin, RetrieveAPIView, DestroyAPIView):
    serializer_class = PageSerializer
    renderer_classes = [
        PlainTextRenderer,
        ImageSVGRenderer,
        JSONRenderer,
        ImageJpegRenderer,
    ]

    def get_queryset(self, *args, **kwargs):
        # This is workaround warning issued when runnnig
        # `./manage.py generateschema`
        # https://github.com/carltongibson/django-filter/issues/966
        if not self.request:
            return Page.objects.none()

        return Page.objects.filter(
            document_version__document__user=self.request.user
        )

    @extend_schema(operation_id="Retrieve")
    def get(self, request, *args, **kwargs):
        """
        Retrieves page resource
        """
        return self.retrieve(request, *args, **kwargs)

    @clock
    def retrieve(self, request, *args, **kwargs):
        """Returns page as json, txt, jpeg or svg image
        """
        instance = self.get_object()
        logger.debug(f"Retrieving page ID={instance.id}")
        # as plain text
        if request.accepted_renderer.format == 'txt':
            data = instance.text
            return Response(data)

        # as html
        if request.accepted_renderer.format in ('html', 'jpeg', 'jpg'):
            logger.debug(f"Page ID={instance.id} requested as html/jpeg/jpg")
            try:
                jpeg_data = instance.get_jpeg()
            except IOError as exc:
                logger.error(exc)
                raise Http404("Jpeg image not available")
            return Response(jpeg_data)

        # as svg (which includes embedded jpeg and HOCRed text overlay)
        if request.accepted_renderer.format == 'svg':
            logger.debug(f"Page ID={instance.id} requested svg")
            content_type = 'image/svg+xml'
            try:
                data = instance.get_svg()
            except IOError:
                # svg not available, try jpeg
                try:
                    data = instance.get_jpeg()
                    content_type = 'image/jpeg'
                except IOError as exc:
                    logger.error(exc)
                    raise Http404("Neither JPEG nor SVG image not available")
            return Response(
                data,
                # either image/jpeg or image/svg+xml
                content_type=content_type
            )

        # by default render page with json serializer
        serializer = self.get_serializer(instance)
        logger.debug(f"Page ID={instance.id} renders default json serializer")

        return Response(serializer.data)

    @extend_schema(operation_id="Single page delete")
    def delete(self, request, *args, **kwargs):
        """
        Deletes page resource
        """
        return self.destroy(request, *args, **kwargs)

    def perform_destroy(self, instance):
        """
        Creates a new document version and copies
        all existing pages to it except current page.
        """
        if instance.is_archived:
            raise APIBadRequest(detail='Deleting archived page is not allowed')

        pages_to_delete = Page.objects.filter(pk__in=[instance.pk])
        old_version = instance.document_version
        doc = instance.document_version.document

        # create new version models i.e. DocumentVersion with Page(s)
        new_version = doc.version_bump(
            page_count=old_version.page_count - 1
        )

        remove_pdf_pages(
            old_version=old_version,
            new_version=new_version,
            page_numbers=[page.number for page in pages_to_delete]
        )

        page_recycle_map = PageRecycleMap(
            total=old_version.page_count,
            deleted=[item.number for item in pages_to_delete]
        )

        page_map = list(page_recycle_map)

        reuse_ocr_data(
            old_version=old_version,
            new_version=new_version,
            page_map=page_map
        )

        reuse_text_field(
            old_version=old_version,
            new_version=new_version,
            page_map=page_map
        )

        page_delete.send(
            sender=Page,
            document_version=new_version
        )


class PagesView(RequireAuthMixin, GenericAPIView):
    serializer_class = PageDeleteSerializer

    @extend_schema(operation_id="Multiple pages delete")
    def delete(self, request):
        serializer = self.serializer_class(data=request.data)

        if serializer.is_valid():
            # delete nodes with specified IDs
            self.delete_pages(page_ids=serializer.data['pages'])
            return Response(
                data=serializer.data,
                status=status.HTTP_204_NO_CONTENT
            )

        return Response(
            serializer.errors,
            status=status.HTTP_400_BAD_REQUEST
        )

    def delete_pages(self, page_ids):
        pages_to_delete = Page.objects.filter(pk__in=page_ids)

        first_page = pages_to_delete.first()

        for page in pages_to_delete:
            if page.is_archived:
                raise APIBadRequest(
                    detail='Deleting archived page is not allowed'
                )

        old_version = first_page.document_version

        count = old_version.pages.count()
        if count <= pages_to_delete.count():
            raise APIBadRequest(
                detail='Document version must have at least one page'
            )

        doc = old_version.document
        new_version = doc.version_bump(
            page_count=old_version.page_count - pages_to_delete.count()
        )

        remove_pdf_pages(
            old_version=old_version,
            new_version=new_version,
            page_numbers=[page.number for page in pages_to_delete]
        )

        page_recycle_map = PageRecycleMap(
            total=old_version.page_count,
            deleted=[item.number for item in pages_to_delete]
        )

        page_map = list(page_recycle_map)

        reuse_ocr_data(
            old_version=old_version,
            new_version=new_version,
            page_map=page_map
        )

        reuse_text_field(
            old_version=old_version,
            new_version=new_version,
            page_map=page_map
        )

        page_delete.send(
            sender=Page,
            document_version=new_version
        )


class PagesReorderView(RequireAuthMixin, GenericAPIView):
    parser_classes = [JSONParser]
    renderer_classes = (JSONRenderer,)
    serializer_class = PagesReorderSerializer

    @extend_schema(operation_id="Reorder")
    def post(self, request):
        """
        Reorders pages within document.
        """
        serializer = self.serializer_class(data=request.data)

        if serializer.is_valid():
            # delete nodes with specified IDs
            self.reorder_pages(pages_data=serializer.data['pages'])
            return Response(
                data=serializer.data,
                status=status.HTTP_204_NO_CONTENT
            )

        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)

    def reorder_pages(self, pages_data):
        pages = Page.objects.filter(
            pk__in=[item['id'] for item in pages_data]
        )
        old_version = pages.first().document_version

        doc = old_version.document
        new_version = doc.version_bump(
            page_count=old_version.page_count
        )

        page_count = old_version.pages.count()

        reorder_pdf_pages(
            old_version=old_version,
            new_version=new_version,
            pages_data=pages_data,
            page_count=page_count
        )

        reordered_list = get_reordered_list(
            pages_data=pages_data,
            page_count=page_count
        )

        reuse_ocr_data(
            old_version=old_version,
            new_version=new_version,
            page_map=list(
                zip(reordered_list, range(1, page_count + 1))
            )
        )

        reuse_text_field(
            old_version=old_version,
            new_version=new_version,
            page_map=list(
                zip(range(1, page_count + 1), reordered_list)
            )
        )

        page_reorder.send(
            sender=Page,
            document_version=new_version
        )


class PagesRotateView(RequireAuthMixin, GenericAPIView):
    parser_classes = [JSONParser]
    renderer_classes = (JSONRenderer,)
    serializer_class = PagesRotateSerializer

    @extend_schema(operation_id="Rotate")
    def post(self, request):
        """
        Rortates one or multiple pages with given angle.
        """
        serializer = self.serializer_class(data=request.data)

        if serializer.is_valid():
            # delete nodes with specified IDs
            self.rotate_pages(pages_data=serializer.data['pages'])
            return Response(
                data=serializer.data,
                status=status.HTTP_204_NO_CONTENT
            )

        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)

    def rotate_pages(self, pages_data):
        pages = Page.objects.filter(
            pk__in=[item['id'] for item in pages_data]
        )
        old_version = pages.first().document_version

        doc = old_version.document
        new_version = doc.version_bump()

        rotate_pdf_pages(
            old_version=old_version,
            new_version=new_version,
            pages_data=annotate_page_data(pages, pages_data, 'angle')
        )

        new_version.generate_previews()

        # page mapping is 1 to 1 as rotation does not
        # add/remove any page
        page_map = [
            (page.number, page.number)
            for page in old_version.pages.all()
        ]

        reuse_text_field(
            old_version=old_version,
            new_version=new_version,
            page_map=page_map
        )

        page_rotate.send(
            sender=Page,
            document_version=new_version
        )


class PagesMoveToFolderView(RequireAuthMixin, GenericAPIView):
    serializer_class = PagesMoveToFolderSerializer
    renderer_classes = (JSONRenderer,)

    @extend_schema(operation_id="Move to folder")
    def post(self, request):
        """
        Moves/extracts one or multiple pages into target folder.

        This operation will create new one or multiple documents (depending
        on ``single_page`` parameter) and place then into target folder.
        ``single_page`` parameter is boolean value which controls whether all
        extracted pages will be placed inside one single document or each
        individual page will be placed into newly created single page document.
        """
        serializer = self.serializer_class(data=request.data)

        if serializer.is_valid():
            self.move_to_folder(serializer.data)
            return Response(
                data=serializer.data,
                status=status.HTTP_204_NO_CONTENT
            )
        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)

    def move_to_folder(self, data):
        pages = Page.objects.filter(pk__in=data['pages'])
        dst_folder = Folder.objects.get(
            pk=data['dst'],
            user=self.request.user
        )
        first_page = pages.first()

        src_old_version = first_page.document_version

        if src_old_version.pages.count() <= 1:
            raise APIBadRequest(
                "Extracting last page of document is not allowed"
            )

        doc = src_old_version.document
        src_new_version = doc.version_bump(
            src_old_version.pages.count() - pages.count()
        )

        remove_pdf_pages(
            old_version=src_old_version,
            new_version=src_new_version,
            page_numbers=[page.number for page in pages]
        )

        page_recycle_map = PageRecycleMap(
            total=src_old_version.page_count,
            deleted=[item.number for item in pages]
        )

        page_map = list(page_recycle_map)

        reuse_ocr_data(
            old_version=src_old_version,
            new_version=src_new_version,
            page_map=page_map
        )
        reuse_text_field(
            old_version=src_old_version,
            new_version=src_new_version,
            page_map=page_map
        )

        if data['single_page']:
            # insert all pages in one single document
            self.move_to_folder_single_paged(
                pages=pages,
                dst_folder=dst_folder,
                title_format=data.get('title_format', None)
            )
        else:
            # there will be one document for each page
            self.move_to_folder_multi_paged(
                pages=pages,
                first_page=first_page,
                dst_folder=dst_folder,
                title_format=data.get('title_format', None)
            )

        page_move_to_folder.send(
            sender=Page,
            document_version=src_new_version
        )

    def move_to_folder_multi_paged(
            self,
            pages,
            first_page,
            dst_folder,
            title_format=None
    ):
        """All extracted pages are inserted into one document"""
        if title_format is None:
            title = f'document-{str(uuid4())}.pdf'
        else:
            title = f'{title_format}.pdf'

        new_doc = Document.objects.create_document(
            title=title,
            lang=first_page.lang,
            user_id=dst_folder.user_id,
            parent=dst_folder,
            ocr_status=OCR_STATUS_SUCCEEDED
        )
        # create new document version which
        # will contain mentioned pages
        dst_version = new_doc.version_bump_from_pages(pages=pages)
        page_map = zip(
            range(1, pages.count() + 1),
            [page.number for page in pages.order_by('number')]
        )
        for src_page, dst_page in zip(
            pages.order_by('number'),
            dst_version.pages.order_by('number'),
        ):
            get_storage_instance().copy_page(
                src=src_page.page_path,
                dst=dst_page.page_path
            )
        reuse_text_field(
            old_version=first_page.document_version,
            new_version=dst_version,
            page_map=page_map
        )

        page_move_to_folder.send(
            sender=Page,
            document_version=dst_version
        )

    def move_to_folder_single_paged(
            self,
            pages,
            dst_folder,
            title_format=None
    ):
        """Each extracted page is inserted into a separate document"""

        pages_count = pages.count()
        for page in pages:
            if title_format is None:
                title = f'page-{str(uuid4())}.pdf'
            else:
                if pages_count > 1:
                    title = f'{title_format}-{str(uuid4())}.pdf'
                else:
                    title = f'{title_format}.pdf'

            doc = Document.objects.create_document(
                title=title,
                lang=page.lang,
                user_id=dst_folder.user_id,
                parent=dst_folder,
                ocr_status=OCR_STATUS_SUCCEEDED
            )
            # create new document version
            # with one page
            doc_version = doc.version_bump_from_pages(pages=[page])
            get_storage_instance().copy_page(
                src=page.page_path,
                dst=doc_version.pages.first().page_path
            )
            reuse_text_field(
                old_version=page.document_version,
                new_version=doc_version,
                page_map=[(1, page.number)]
            )

            page_move_to_folder.send(
                sender=Page,
                document_version=doc_version
            )


class PagesMoveToDocumentView(RequireAuthMixin, GenericAPIView):
    serializer_class = PagesMoveToDocumentSerializer
    renderer_classes = (JSONRenderer,)

    @extend_schema(operation_id="Move to document")
    def post(self, request):
        """
        Moves one or multiple pages from source document to target document.

        Both source and target documents' version will be incremented
        by one.
        """
        serializer = self.serializer_class(data=request.data)

        if serializer.is_valid():
            if serializer.data.get('merge', False):
                self.merge_to_document(serializer.data)
            else:
                self.move_to_document(serializer.data)
            return Response(
                data=serializer.data,
                status=status.HTTP_204_NO_CONTENT
            )

        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)

    def merge_to_document(self, data):
        """Merge creates target document version ONLY from source pages

        There is a total merge and partial merge.
        Total merge is when ALL pages of the source are involved. During total
        merge source document is deleted.
        When not all pages are involved in the merge - we say it is a partial
        merge. During partial merge source documents is NOT deleted.
        """
        pages = Page.objects.filter(
            pk__in=data['pages']
        )
        dst_document = Document.objects.get(
            pk=data['dst'],
            user=self.request.user
        )
        src_old_version = pages.first().document_version
        doc = src_old_version.document
        pages_count = pages.count()

        if src_old_version.pages.count() == pages_count:
            # destination new version will have same
            # number of pages as source document count
            dst_new_version = dst_document.version_bump(
                page_count=pages_count,
                short_description=f'{pages_count} page(s) merged in'
            )
            total_merge(
                src_old_version=src_old_version,
                dst_new_version=dst_new_version
            )
            page_move_to_document.send(
                sender=Page,
                document_version=dst_new_version
            )
        else:
            src_new_version = doc.version_bump(
                page_count=src_old_version.pages.count() - pages_count,
                short_description=f'{pages_count} page(s) merged out'
            )
            dst_new_version = dst_document.version_bump(
                page_count=pages_count,
                short_description=f'{pages_count} page(s) merged in'
            )
            partial_merge(
                src_old_version=src_old_version,
                src_new_version=src_new_version,
                dst_new_version=dst_new_version,
                page_numbers=[p.number for p in pages.order_by('number')]
            )
            page_move_to_document.send(
                sender=Page,
                document_version=src_new_version
            )
            page_move_to_document.send(
                sender=Page,
                document_version=dst_new_version
            )

    def move_to_document(self, data):
        pages = Page.objects.filter(
            pk__in=data['pages']
        )
        dst_document = Document.objects.get(
            pk=data['dst'],
            user=self.request.user
        )
        src_old_version = pages.first().document_version
        dst_old_version = dst_document.versions.last()
        pages_count = pages.count()
        position = data['position']
        if position < 0:
            position = dst_old_version.pages.count()

        doc = src_old_version.document
        src_new_version = doc.version_bump(
            page_count=src_old_version.pages.count() - pages_count,
            short_description=f'{pages_count} page(s) moved out'
        )
        dst_new_version = dst_document.version_bump(
            page_count=dst_old_version.pages.count() + pages_count,
            short_description=f'{pages_count} page(s) moved in'
        )

        remove_pdf_pages(
            old_version=src_old_version,
            new_version=src_new_version,
            page_numbers=[page.number for page in pages]
        )

        page_recycle_map = PageRecycleMap(
            total=src_old_version.page_count,
            deleted=[item.number for item in pages]
        )

        page_map = list(page_recycle_map)

        reuse_ocr_data(
            old_version=src_old_version,
            new_version=src_new_version,
            page_map=page_map
        )

        reuse_text_field(
            old_version=src_old_version,
            new_version=src_new_version,
            page_map=page_map
        )

        insert_pdf_pages(
            src_old_version=src_old_version,
            dst_old_version=dst_old_version,
            dst_new_version=dst_new_version,
            src_page_numbers=[p.number for p in pages.order_by('number')],
            dst_position=position
        )

        reuse_ocr_data_multi(
            src_old_version=src_old_version,
            dst_old_version=dst_old_version,
            dst_new_version=dst_new_version,
            position=position,
            page_numbers=[page.number for page in pages]
        )

        reuse_text_field_multi(
            src_old_version=src_old_version,
            dst_old_version=dst_old_version,
            dst_new_version=dst_new_version,
            position=position,
            page_numbers=[page.number for page in pages]
        )

        page_move_to_document.send(
            sender=Page,
            document_version=src_new_version
        )

        page_move_to_document.send(
            sender=Page,
            document_version=dst_new_version
        )



from asgiref.sync import async_to_sync


class RequireAuth:

    def connect(self):
        self.user = self.scope["user"]
        if self.user is None:
            return self.close()

        async_to_sync(
            self.channel_layer.group_add
        )(self.group_name, self.channel_name)
        # When websocket clients sends non empty 'Sec-WebSocket-Protocol'
        # Chrome browser expects response from the server with non empty
        #  Sec-WebSocket-Protocol as well.
        # -> self.accept() sends empty Sec-WebSocket-Protocol WebSocket()
        #   function fails with an error
        # -> self.accept('access_token') send NON EMPTY Sec-WebSocket-Protocol
        #   header with value 'access_token'
        # which fixes the error in Chrome Browser
        # https://github.com/django/channels/issues/1369#issuecomment-724299511
        if 'access_token' in self.scope['subprotocols']:
            self.accept('access_token')
        else:
            self.accept()



from haystack import indexes
from papermerge.core.models import Document, Folder


class DocumentIndex(indexes.SearchIndex, indexes.Indexable):
    indexed_content = indexes.CharField(document=True, use_template=True)
    id = indexes.CharField(model_attr='id')
    user = indexes.CharField(model_attr='document__user')
    title = indexes.CharField(model_attr='title')
    last_version_text = indexes.CharField()
    text = indexes.CharField()  # alias for `last_version_text`
    tags = indexes.MultiValueField()
    highlight = indexes.CharField()
    breadcrumb = indexes.MultiValueField()
    node_type = indexes.CharField()

    def prepare_last_version_text(self, obj):
        last_document_version = obj.versions.last()
        if last_document_version:
            return last_document_version.text

        return ''

    def prepare_breadcrumb(self, instance):
        list_of_titles = [
            item.title for item in instance.get_ancestors(include_self=False)
        ]

        return list_of_titles

    def prepare_node_type(self, obj):
        return 'document'

    def prepare_text(self, obj):
        return self.prepare_last_version_text(obj)

    def prepare_tags(self, obj):
        return [tag.name for tag in obj.tags.all()]

    def get_model(self):
        return Document


class FolderIndex(indexes.SearchIndex, indexes.Indexable):
    indexed_content = indexes.CharField(document=True, use_template=True)
    id = indexes.CharField(model_attr='id')
    title = indexes.CharField(model_attr='title')
    user = indexes.CharField(model_attr='user')
    node_type = indexes.CharField()
    breadcrumb = indexes.MultiValueField()
    tags = indexes.MultiValueField()

    def prepare_breadcrumb(self, instance):
        breadcrumb_items = [
            item.title
            for item in instance.get_ancestors(include_self=False)
        ]
        return breadcrumb_items

    def prepare_node_type(self, obj):
        return 'folder'

    def prepare_tags(self, obj):
        return [tag.name for tag in obj.tags.all()]

    def get_model(self):
        return Folder



import logging
import os
from typing import Optional
from os.path import getsize
from pikepdf import Pdf

from django.db import models
from django.db import transaction

from papermerge.core.lib.path import DocumentPath, PagePath
from papermerge.core.signal_definitions import document_post_upload
from papermerge.core.storage import get_storage_instance, abs_path

from .node import BaseTreeNode
from .utils import (
    OCR_STATUS_SUCCEEDED,
    OCR_STATUS_UNKNOWN,
    OCR_STATUS_CHOICES
)
from .page import Page
from .document_version import DocumentVersion


logger = logging.getLogger(__name__)


class UploadStrategy:
    """
    Defines how to proceed with uploaded file
    """
    # INCREMENT - Uploaded file is inserted into the newly created
    #   document version
    INCREMENT = 1
    # MERGE - Uploaded file is merged with last file version
    #   and inserted into the newly created document version
    MERGE = 2


class DocumentManager(models.Manager):

    @transaction.atomic
    def create_document(
        self,
        user_id,
        title,
        lang,
        size=0,
        page_count=0,
        file_name=None,
        parent=None,
        **kwargs
    ):
        doc = Document(
            title=title,
            lang=lang,
            user_id=user_id,
            parent=parent,
            **kwargs
        )
        doc.save()

        document_version = DocumentVersion(
            document=doc,
            number=1,  # versioning number starts with 1
            file_name=file_name,
            size=0,
            page_count=0,
            short_description="Original"
        )
        document_version.save()
        # Important! - first document must inherit metakeys from
        # parent folder
        # if parent:
        #    doc.inherit_kv_from(parent)
        return doc

    def _get_parent(self, parent_id):
        """
        Returns parent node instance based on ``parent_id``
        """
        parent = None

        if parent_id is None or parent_id == '':
            parent = None
        else:
            try:
                parent = BaseTreeNode.objects.get(id=parent_id)
            except BaseTreeNode.DoesNotExist:
                parent = None

        return parent


class DocumentQuerySet(models.QuerySet):
    pass


CustomDocumentManager = DocumentManager.from_queryset(DocumentQuerySet)


class Document(BaseTreeNode):

    # Will this document be OCRed?
    # If True this document will be OCRed
    # If False, OCR operation will be skipped for this document
    ocr = models.BooleanField(default=True)

    # This field is updated by
    # `papermerge.avenues.consumers.document.DocumentConsumer`.
    #
    # Can be one of: 'unknown', 'received', 'started',
    # 'failed', 'succeeded' - these values correspond to
    # celery's task statuses
    ocr_status = models.CharField(
        choices=OCR_STATUS_CHOICES,
        default=OCR_STATUS_UNKNOWN,
        max_length=32
    )

    objects = CustomDocumentManager()

    @property
    def idified_title(self):
        """
        Returns a title with ID part inserted before extention

        Example:
            input: title="invoice.pdf", id="233453"
            output: invoice-233453.pdf
        """
        base_title_arr = self.title.split('.')[:-1]
        base_title = '.'.join(base_title_arr)
        ext = self.title.split('.')[-1]

        return f'{base_title}-{self.id}.{ext}'

    class Meta:
        verbose_name = "Document"
        verbose_name_plural = "Documents"

    def upload(
            self,
            payload,
            file_path,
            file_name,
            strategy=UploadStrategy.INCREMENT
    ):
        """
        Associates payload with specific document version.

        If document has zero sized document version, it will associate
        payload with that (existing) version, otherwise it will create
        new document version and associate it the payload.
        """
        pdf = Pdf.open(payload)

        document_version = self.versions.filter(size=0).last()

        if not document_version:
            document_version = DocumentVersion(
                document=self,
                number=self.versions.count(),
                lang=self.lang
            )

        document_version.file_name = file_name
        document_version.size = getsize(file_path)
        document_version.page_count = len(pdf.pages)

        get_storage_instance().copy_doc(
            src=file_path,
            dst=document_version.document_path
        )

        document_version.save()
        document_version.create_pages()
        pdf.close()

        document_post_upload.send(
            sender=self.__class__,
            document_version=document_version
        )

    def version_bump_from_pages(self, pages):
        """
        Creates new version for the document.
        PDF pages in the newly create document version is copied
        from ``pages``.

        ``pages`` is an iterable. Maybe either Page queryset or
        a list of ``Page`` model instances. Used when extracting/moving
        pages from the document to folder (thus creating new document).
        """
        if isinstance(pages, list):
            first_page = pages[0]
            page_count = len(pages)
        else:  # pages queryset
            first_page = pages.first()
            page_count = pages.count()
        source_pdf = Pdf.open(
            abs_path(first_page.document_version.document_path.url)
        )
        dst_pdf = Pdf.new()

        document_version = self.versions.filter(size=0).last()

        if not document_version:
            document_version = DocumentVersion(
                document=self,
                number=self.versions.count(),
                lang=self.lang
            )

        for page in pages:
            pdf_page = source_pdf.pages.p(page.number)
            dst_pdf.pages.append(pdf_page)

        document_version.file_name = first_page.document_version.file_name
        document_version.page_count = page_count
        document_version.save()

        dirname = os.path.dirname(
            abs_path(document_version.document_path.url)
        )
        os.makedirs(dirname, exist_ok=True)

        dst_pdf.save(abs_path(document_version.document_path.url))

        document_version.size = getsize(
            abs_path(document_version.document_path.url)
        )
        document_version.save()

        document_version.create_pages()
        source_pdf.close()
        dst_pdf.close()

        return document_version

    def version_bump(
        self,
        page_count=None,
        short_description: Optional[str] = ''
    ):
        """
        Increment document's version.

        Creates new document version (old version = old version + 1) and
        copies all attributes from current document version.
        If ``page_count`` is not None new document version will
        have ``page_count`` pages (useful when page was deleted or number of
        new pages were merged into the document).
        If ``page_count`` is None new version will have same number of pages as
        previous document (useful when new document was OCRed or
        when pages were rotated)
        """
        last_doc_version = self.versions.last()
        new_page_count = last_doc_version.page_count
        if page_count:
            new_page_count = page_count

        new_doc_version = DocumentVersion(
            document=self,
            number=last_doc_version.number + 1,
            file_name=last_doc_version.file_name,
            size=0,  # TODO: set to newly created file size
            page_count=new_page_count,
            short_description=short_description,
            lang=last_doc_version.lang
        )
        new_doc_version.save()

        for page_number in range(1, new_page_count + 1):
            Page.objects.create(
                document_version=new_doc_version,
                number=page_number,
                page_count=new_page_count,
                lang=last_doc_version.lang
            )

        return new_doc_version

    def __repr__(self):
        return f"Document(id={self.pk}, title={self.title})"

    def __str__(self):
        return self.title

    @property
    def file_ext(self):
        _, ext = os.path.splitext(self.file_name)
        return ext

    @property
    def absfilepath(self):
        return abs_path(
            self.path().url()
        )

    def path(self, version=None):

        if version is None:
            version = self.version

        version = int(version)

        result = DocumentPath(
            user_id=self.user.id,
            document_id=self.id,
            version=version,
            file_name=self.file_name,
        )

        return result

    def page_paths(self, version=None):
        """
        Enables document instance to get quickly page
        paths:

            page_path = doc.page_path[2]
            page_path.url() # local url to second page of the doc.

        This is shortcut method when most used Step(1) is required.
        """

        results = [None]  # indexing starts from 1

        page_count = self.get_pagecount(version=version)

        for page_num in range(1, page_count + 1):
            page_path = PagePath(
                document_path=self.path(version=version),
                page_num=page_num,
                page_count=self.get_pagecount(version=version)
            )
            results.append(page_path)

        return results

    def get_page_path(self, page_num, version=None):
        return PagePath(
            document_path=self.path(version=version),
            page_num=page_num,
            page_count=self.page_count
        )

    def preview_path(self, page, size=None):

        if page > self.page_count or page < 0:
            raise ValueError("Page index out of bound")

        file_name = os.path.basename(self.file_name)
        root, _ = os.path.splitext(file_name)
        page_count = self.pages_num

        if not size:
            size = "orig"

        if page_count <= 9:
            fmt_page = "{root}-page-{num:d}.{ext}"
        elif page_count > 9 and page_count < 100:
            fmt_page = "{root}-page-{num:02d}.{ext}"
        elif page_count > 100:
            fmt_page = "{root}-page-{num:003d}.{ext}"

        return os.path.join(
            self.dir_path,
            str(size),
            fmt_page.format(
                root=root, num=int(page), ext="jpg"
            )
        )

    @property
    def name(self):
        root, ext = os.path.splitext(self.file_name)
        return root

    def add_tags(self, tags):
        """
        tags is an iteratable of papermerge.core.models.Tag instances
        """
        for tag in tags:
            self.tags.add(
                tag,
                tag_kwargs={'user': self.user}
            )

    def get_ocr_status(self):
        """
        Returns OCR status of the document.

        Document model knows only limited information about
        document OCR status. From point of view of the document
        OCR status can be one of following:

            * succeeded - when document.text field is non empty
            * unknown - when document.text is empty

        In case of "unknown" OCR status application will need to query
        different parts of the system to figure out more details
        about OCR status.
        """
        if len(self.text) > 0:
            return OCR_STATUS_SUCCEEDED

        return OCR_STATUS_UNKNOWN



from rest_framework.generics import GenericAPIView
from rest_framework.response import Response
from rest_framework.generics import get_object_or_404

from papermerge.core.models import Document
from papermerge.core.tasks import (
    ocr_document_task,
    post_ocr_document_task
)
from papermerge.core.storage import default_storage
from papermerge.core.serializers import OcrSerializer

from .mixins import RequireAuthMixin


class OCRView(RequireAuthMixin, GenericAPIView):

    serializer_class = OcrSerializer

    def post(self, request):
        """
        Starts OCR for document version
        """
        doc_id = request.data['doc_id']
        lang = request.data['lang']
        doc = get_object_or_404(Document.objects, pk=doc_id)
        namespace = getattr(default_storage, 'namespace', None)

        ocr_document_task.apply_async(
            kwargs={
                'document_id': str(doc.id),
                'lang': lang,
                'namespace': namespace,
                'user_id': str(request.user.id)
            },
            link=post_ocr_document_task.s(namespace)
        )

        return Response({"message": "OCR successfully started"})



from .content_types import ContentTypeRetrieve
from .tags import TagsViewSet
from .groups import GroupsViewSet
from .permissions import PermissionsList
from .users import (
    UsersViewSet,
    UserChangePassword,
    CurrentUserView
)
from .nodes import (
    NodesViewSet,
    NodesMoveView,
    NodesDownloadView,
    InboxCountView,
    NodeTagsView,
)
from .documents import (
    DocumentUploadView,
    DocumentsMergeView,
    DocumentOcrTextView
)
from .document_versions import (
    DocumentVersionsDownloadView,
    DocumentVersionView
)
from .documents import DocumentDetailsViewSet
from .folders import FoldersViewSet
from .pages import (
    PageView,
    PagesView,
    PagesReorderView,
    PagesRotateView,
    PagesMoveToFolderView,
    PagesMoveToDocumentView
)
from .tasks import OCRView
from .preferences import CustomUserPreferencesViewSet
from .login import LoginView
from .logout import LogoutView, LogoutAllView
from .tokens import TokensViewSet
from .version import VersionView



from django.contrib.auth.models import Permission

from rest_framework_json_api import serializers

from .content_type import ContentTypeSerializer


class PermissionSerializer(serializers.ModelSerializer):

    included_serializers = {
        'content_type': ContentTypeSerializer,
    }

    class Meta:
        model = Permission
        resource_name = 'permissions'
        fields = (
            'id',
            'name',
            'codename',
            'content_type',
        )
        read_only_fields = (
            'id',
            'name',
            'codename',
            'content_type'
        )

    class JSONAPIMeta:
        included_resources = ['content_type']



import os
import logging

import ocrmypdf

from papermerge.core.storage import abs_path
from papermerge.core.lib import mime
from papermerge.core.lib.tiff import convert_tiff2pdf
from papermerge.core.lib.path import (
    DocumentPath,
)


logger = logging.getLogger(__name__)

STARTED = "started"
COMPLETE = "complete"


def notify_hocr_ready(page_path, **kwargs):
    pass


def notify_txt_ready(page_path, **kwargs):
    pass


def notify_pre_page_ocr(page_path, **kwargs):
    pass


def _ocr_document(
    input_doc_path: DocumentPath,
    target_doc_path,
    lang,
    preview_width,
):

    # file_name = kwargs.pop('file_name', None)

    # if not file_name:
    #    input_file_name = input_doc_path.file_name

    sidecars_dir = abs_path(target_doc_path.dirname_sidecars())

    input_document = abs_path(input_doc_path.path)

    output_document = abs_path(target_doc_path.path)

    output_dir = os.path.dirname(output_document)

    if not os.path.exists(output_dir):
        os.makedirs(
            output_dir,
            exist_ok=True
        )

    ocrmypdf.ocr(
        input_document,
        output_document,
        lang=lang,
        plugins=["ocrmypdf_papermerge.plugin"],
        progress_bar=False,
        output_type='pdf',
        pdf_renderer='hocr',
        use_threads=True,
        force_ocr=True,
        keep_temporary_files=False,
        sidecar_dir=sidecars_dir,
        sidecar_format='svg',
        preview_width=preview_width,
        deskew=True
    )


def ocr_document(
    user_id,
    document_id,
    file_name,
    lang,
    version,
    target_version,
    namespace='',
):
    lang = lang.lower()
    doc_path = DocumentPath(
        user_id=user_id,
        document_id=document_id,
        file_name=file_name,
        version=version
    )
    target_doc_path = DocumentPath.copy_from(
        doc_path,
        version=target_version
    )

    mime_type = mime.Mime(
        abs_path(doc_path.url)
    )

    if mime_type.is_pdf() or mime_type.is_image():
        _ocr_document(
            input_doc_path=doc_path,
            target_doc_path=target_doc_path,
            lang=lang,
            preview_width=300
        )
    elif mime_type.is_tiff():
        new_filename = convert_tiff2pdf(
            doc_url=abs_path(doc_path.url)
        )
        # now .pdf
        orig_file_name = doc_path.file_name
        doc_path.file_name = new_filename
        # and continue as usual
        _ocr_document(
            doc_path=doc_path,
            lang=lang,
            user_id=user_id,
            document_id=document_id,
            # Pass original file_name i.e. tiff file name as well.
            file_name=orig_file_name,
            namespace=namespace,
            version=version
        )
    else:
        logger.error(
            f" user_id={user_id}"
            f" doc_id={document_id}"
        )
        return True

    return True



from django.contrib.auth.models import Group
from rest_framework_json_api.views import ModelViewSet
from rest_framework_json_api.renderers import JSONRenderer
from rest_framework.response import Response
from drf_spectacular.utils import extend_schema

from papermerge.core.serializers.group import (
    GroupSerializer,
    Data_GroupSerializer
)
from papermerge.core.auth import CustomModelPermissions
from .mixins import RequireAuthMixin


class GroupsViewSet(RequireAuthMixin, ModelViewSet):
    """
    Group endpoint
    """
    queryset = Group.objects.all()
    serializer_class = GroupSerializer
    renderer_classes = (JSONRenderer,)
    permission_classes = [CustomModelPermissions]

    @extend_schema(
        responses={201: Data_GroupSerializer},
    )
    def retrieve(self, request, *args, **kwargs):
        instance = self.get_object()
        serializer = self.get_serializer(instance)
        return Response(serializer.data)



import yaml
import logging.config

from corsheaders.defaults import default_headers as default_cors_headers
from configula import Configula
from papermerge.core.openapi.append import JSONAPI_COMPONENTS
from papermerge.core.version import __version__

logger = logging.getLogger(__name__)

config = Configula()

ALLOWED_HOSTS = config.get(
    'main',
    'allowed_hosts',
    default=['*']
)

redis_host = config.get('redis', 'host', default=None)
redis_port = config.get('redis', 'port', default=None)

if redis_host and redis_port:
    CELERY_BROKER_URL = f"redis://{redis_host}:{redis_port}/0"
else:
    CELERY_BROKER_URL = 'memory://localhost/'

CELERY_WORKER_HIJACK_ROOT_LOGGER = False
CELERY_ACCEPT_CONTENT = ['json']
CELERY_TASK_SERIALIZER = 'json'
CELERY_RESULT_SERIALIZER = 'json'
CELERY_TASK_CREATE_MISSING_QUEUES = True
CELERY_TASK_DEFAULT_EXCHANGE = 'papermerge'
CELERY_TASK_DEFAULT_EXCHANGE_TYPE = 'direct'
CELERY_TASK_DEFAULT_ROUTING_KEY = 'papermerge'

if redis_host and redis_port:
    CHANNEL_LAYERS = {
        'default': {
            'BACKEND': 'channels_redis.core.RedisChannelLayer',
            'CONFIG': {
                "hosts": [(redis_host, redis_port)],
            },
        },
    }

DEBUG = config.get('main', 'debug', False)
PAPERMERGE_NAMESPACE = config.get('main', 'namespace', None)

DEFAULT_AUTO_FIELD = 'django.db.models.AutoField'

# Custom signal processor handles connections errors (to elasticsearch)
# and reports them as warnings. This way, even when no connection to ES
# is available, documents, folders, pages etc can still be used
ELASTICSEARCH_DSL_SIGNAL_PROCESSOR = 'papermerge.search.signals.CustomSignalProcessor'  # noqa

MEDIA_URL = config.get(
    'media',
    'url',
    default='/media/'
)

SECRET_KEY = config.get('main', 'secret_key')

SITE_ID = 1

STATIC_URL = config.get(
    'static',
    'dir',
    default='/static/'
)

PAPERMERGE_MIMETYPES = [
    'application/octet-stream',
    'application/pdf',
    'image/png',
    'image/jpeg',
    'image/jpg',
    'image/tiff'
]

# For each user create special folders
# i.e. create ".inbox" and ".home" folders
PAPERMERGE_CREATE_SPECIAL_FOLDERS = True

# Application definition

INSTALLED_APPS = [
    'rest_framework',
    'rest_framework.authtoken',
    'knox',
    'rest_framework_json_api',
    'corsheaders',
    'drf_spectacular',
    'drf_spectacular_sidecar',
    'django.contrib.auth',
    'django.contrib.sites',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'papermerge.core.apps.CoreConfig',
    'papermerge.search.apps.SearchConfig',
    'papermerge.notifications.apps.NotificationsConfig',
    'django.contrib.contenttypes',
    'dynamic_preferences',
    'dynamic_preferences.users.apps.UserPreferencesConfig',
    'channels',
    'haystack',
]

# include elasticsearch apps only if PAPERMERGE_ELASTICSEARCH_HOSTS
# and PAPERMERGE_ELASTICSEARCH_PORT are defined
# and have non-empty value

MIDDLEWARE = [
    'corsheaders.middleware.CorsMiddleware',
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

ROOT_URLCONF = 'config.urls'
AUTH_USER_MODEL = 'core.User'
WSGI_APPLICATION = 'config.wsgi.application'
ASGI_APPLICATION = 'config.asgi.application'


TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

if config.has_mysql:
    # Requires MySQL > 5.7.7 or innodb_large_prefix set to on
    SILENCED_SYSTEM_CHECKS = ['mysql.E001']

FILE_UPLOAD_HANDLERS = [
    'django.core.files.uploadhandler.TemporaryFileUploadHandler'
]

STATICFILES_FINDERS = [
    'django.contrib.staticfiles.finders.FileSystemFinder',
    'django.contrib.staticfiles.finders.AppDirectoriesFinder',
]

LANGUAGES = [
    ('de', 'Deutsch'),
    ('en', 'English'),
    ('fr', 'Français'),
]
TIME_ZONE = config.get('main', 'timezone', default='Europe/Berlin')
USE_I18N = True
USE_L10N = True
USE_TZ = True
LANGUAGE_CODE = config.get(
    'main',
    'language_code',
    default='en'
)

DATE_FORMAT = '%d/%m/%Y'
DATE_INPUT_FORMATS = ['%d/%m/%Y']

# Password validation
# https://docs.djangoproject.com/en/3.2/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]

REST_FRAMEWORK = {
    'DEFAULT_AUTHENTICATION_CLASSES': [
        # knox token based authentication
        'knox.auth.TokenAuthentication',
    ],
    'PAGE_SIZE': 10,
    'EXCEPTION_HANDLER': 'rest_framework_json_api.exceptions.exception_handler',
    'DEFAULT_PAGINATION_CLASS':
        'papermerge.core.openapi.pagination.JsonApiPagination',
    'DEFAULT_PARSER_CLASSES': (
        'rest_framework_json_api.parsers.JSONParser',
        'rest_framework.parsers.JSONParser',
        'rest_framework.parsers.MultiPartParser'
    ),
    'DEFAULT_RENDERER_CLASSES': (
        'rest_framework_json_api.renderers.JSONRenderer',
        'rest_framework.renderers.JSONRenderer',
    ),
    'DEFAULT_METADATA_CLASS': 'rest_framework_json_api.metadata.JSONAPIMetadata',
    'DEFAULT_SCHEMA_CLASS': 'drf_spectacular.openapi.AutoSchema',
    'DEFAULT_FILTER_BACKENDS': (
        'rest_framework_json_api.filters.OrderingFilter',
        'rest_framework_json_api.django_filters.DjangoFilterBackend',
        'rest_framework.filters.SearchFilter',
    ),
    'SEARCH_PARAM': 'filter[search]',
    'TEST_REQUEST_RENDERER_CLASSES': (
        'rest_framework_json_api.renderers.JSONRenderer',
    ),
    'TEST_REQUEST_DEFAULT_FORMAT': 'vnd.api+json'
}

REST_KNOX = {
  # Setting the TOKEN_TTL to None will create tokens that never expire.
  'TOKEN_TTL': None,
}

LOGGING_CFG_FILENAME = config.get('main', 'logging_cfg', None)
if LOGGING_CFG_FILENAME:
    dict_config = yaml.load(open(LOGGING_CFG_FILENAME), Loader=yaml.FullLoader)
    logging.config.dictConfig(dict_config)
else:
    LOGGING = {
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'verbose': {
                'format': '%(levelname)s %(asctime)s %(module)s %(message)s'
            },
        },
        'handlers': {
            'console': {
                'class': 'logging.StreamHandler',
                'formatter': 'verbose'
            },
        },
        'root': {
            'level': 'INFO',
            'handlers': ['console', ],
        },
        'loggers': {
            'django.db.backends': {
                'level': 'INFO',
                'handlers': ['console', ],
                'propagate': False,
            },
            'django.security.DisallowedHost': {
                'level': 'INFO',
                'handlers': ['console', ],
                'propagate': False,
            },
            'papermerge.core.tasks': {
                'level': 'DEBUG',
                'handlers': ['console', ],
                'propagate': True,
            },
            'papermerge.search.tasks': {
                'level': 'DEBUG',
                'handlers': ['console', ],
                'propagate': True,
            },
            'celery': {
                'level': 'DEBUG',
                'handlers': ['console', ],
                'propagate': True,
            },
        },
    }

CORS_ALLOW_HEADERS = list(default_cors_headers) + [
    "Authorization",
    "Content-Disposition",
]

CORS_EXPOSE_HEADERS = ["Content-Disposition"]

CORS_ALLOW_ALL_ORIGINS = True


SPECTACULAR_SETTINGS = {
    'TITLE': 'Papermerge REST API',
    'DESCRIPTION': 'Document management system designed for digital archives',
    'VERSION': __version__,
    'APPEND_COMPONENTS': JSONAPI_COMPONENTS
}

SEARCH_ENGINES_MAP = {
    'elastic': 'haystack.backends.elasticsearch7_backend.Elasticsearch7SearchEngine',
    'elastic7': 'haystack.backends.elasticsearch7_backend.Elasticsearch7SearchEngine',
    'elasticsearch7': 'haystack.backends.elasticsearch7_backend.Elasticsearch7SearchEngine',
    'elasticsearch': 'haystack.backends.elasticsearch7_backend.Elasticsearch7SearchEngine',
    'es7': 'haystack.backends.elasticsearch7_backend.Elasticsearch7SearchEngine',
    'es': 'haystack.backends.elasticsearch7_backend.Elasticsearch7SearchEngine',
    'solr': 'haystack.backends.solr_backend.SolrEngine',
    'whoosh': 'haystack.backends.whoosh_backend.WhooshEngine',
    'xapian': 'xapian_backend.XapianEngine',
}

HAYSTACK_DOCUMENT_FIELD = 'indexed_content'

search_engine = config.get('search', 'engine', default='xapian')


HAYSTACK_CONNECTIONS = {
    'default': {
        'ENGINE': SEARCH_ENGINES_MAP[search_engine],
    },
}

HAYSTACK_SIGNAL_PROCESSOR = 'papermerge.search.signals.SignalProcessor'



from django.db import models
from haystack import signals

from papermerge.core.models import (
    DocumentVersion,
    Document,
    Folder,
    BaseTreeNode
)
from papermerge.search.tasks import update_index


class SignalProcessor(signals.BaseSignalProcessor):
    def setup(self):
        for klass in (DocumentVersion, Document, Folder, BaseTreeNode):
            models.signals.post_save.connect(
                self.enqueue_save, sender=klass
            )
            models.signals.post_delete.connect(
                self.enqueue_delete, sender=klass
            )

    def teardown(self):
        for klass in (DocumentVersion, Document, Folder, BaseTreeNode):
            models.signals.post_save.disconnect(self.enqueue_save, sender=klass)
            models.signals.post_delete.disconnect(
                self.enqueue_delete,
                sender=klass
            )

    def enqueue_save(self, sender, instance, **kwargs):
        return update_index.apply_async()

    def enqueue_delete(self, sender, instance, **kwargs):
        return update_index.apply_async()



import logging

from django.core.exceptions import PermissionDenied

from rest_framework.generics import GenericAPIView
from rest_framework import status
from rest_framework.response import Response
from rest_framework.parsers import JSONParser
from rest_framework.renderers import JSONRenderer
from rest_framework_json_api.views import ModelViewSet
from rest_framework_json_api.renderers import JSONRenderer as JSONAPIRenderer
from rest_framework_json_api.parsers import JSONParser as JSONAPIParser

from drf_spectacular.utils import extend_schema

from papermerge.core.serializers.user import data_user_response_example
from papermerge.core.serializers import (
    Data_UserSerializer,
    UserSerializer,
    PasswordSerializer
)
from papermerge.core.models import User
from papermerge.core.auth import CustomModelPermissions
from .mixins import RequireAuthMixin

logger = logging.getLogger(__name__)


class UsersViewSet(RequireAuthMixin, ModelViewSet):
    """
    Users endpoint
    """
    queryset = User.objects.all()
    serializer_class = UserSerializer
    renderer_classes = (JSONAPIRenderer,)
    parser_classes = [JSONAPIParser]
    permission_classes = [CustomModelPermissions]


class UserChangePassword(RequireAuthMixin, GenericAPIView):
    parser_classes = [JSONParser]
    renderer_classes = (JSONRenderer,)
    serializer_class = PasswordSerializer
    http_method_names = ['post', 'head', 'options']

    def post(self, request, pk):
        """
        Change password of the user identified with UUID/pk in the URL
        """
        if not self._has_perm(request, pk):
            raise PermissionDenied

        serializer = PasswordSerializer(data=request.data)
        user = User.objects.get(pk=pk)
        if serializer.is_valid():
            user.set_password(serializer.validated_data['password'])
            user.save()
            return Response({'status': 'password set'})
        else:
            return Response(
                serializer.errors,
                status=status.HTTP_400_BAD_REQUEST
            )

    def _has_perm(self, request, pk):
        """
        Does user have permission to perform 'change user password'?

        Same endpoint is used when changing other users password and your
        own user password. When changing own user password, user does not
        need to have 'change_user' permission.
        """
        if request.user.pk == pk:
            return True

        # here changing password for other users
        if request.user.has_perm('core.change_user'):
            return True


class CurrentUserView(RequireAuthMixin, GenericAPIView):
    resource_name = 'users'
    serializer_class = UserSerializer
    renderer_classes = (JSONAPIRenderer,)

    @extend_schema(
        responses={200: Data_UserSerializer},
        examples=[data_user_response_example]
    )
    def get(self, request):
        """
        Retrieves current user details.

        Detailed information includes, among others, information about home
        folder, inbox folder and permission of the current user.

        Note that this endpoint returns user permissions i.e. what actions
        current user is authorized to perform.
        """
        serializer = UserSerializer(self.request.user)
        return Response(serializer.data)

    def get_queryset(self):
        # This is workaround warning issued when runnnig
        # `./manage.py generateschema`
        # https://github.com/carltongibson/django-filter/issues/966
        if not self.request:
            return User.objects.none()

        return super().get_queryset()



from pytz import common_timezones

from django.conf import settings

from dynamic_preferences.preferences import Section as OrigSection
from dynamic_preferences.types import ChoicePreference
from dynamic_preferences.users.registries import user_preferences_registry

from .lib.lang import get_ocr_langs, get_default_ocr_lang


def _get_timezone_choices():
    return list((tz, tz) for tz in common_timezones)


class Section(OrigSection):
    """
    Section has following attributes:

        * name
        * verbose_name
        * help_text
        * icon_name
        * visible (default=True)

    If Section has visibility attribute set to False - it won't be
    displayed/visible in User preferences view.
    Example:
    if administrator decides to disable email_routing feature (i.e. he/she
    preferes all incoming email to land in superuser Inbox) with:

        PAPERMERGE_IMPORT_MAIL_BY_USER = False
        PAPERMERGE_IMPORT_MAIL_BY_SECRET = False

    Then, there is no point (it is even confusing) for user to fiddle with
    Email Routing options. From UX poing of view - it is better not to display
    Email Routing section at all.
    """

    def __init__(
        self,
        name,
        verbose_name=None,
        help_text=None,
        icon_name=None,
        visible=True,
    ):
        super().__init__(
            name=name,
            verbose_name=verbose_name
        )
        self.help_text = help_text
        self.icon_name = icon_name
        self.visible = visible


localization = Section(
    'localization',
    verbose_name="Localization",
    icon_name='globe-americas',
    help_text='Timezone, date and time formats'
)

ocr = Section(
    'ocr',
    verbose_name='Opical Character Recognition',
    icon_name='eye',
    help_text='Choose default OCR Language'
)


@user_preferences_registry.register
class TimezoneGlobal(ChoicePreference):
    help_text = """
    Timezone
"""
    section = localization
    name = "timezone"
    choices = _get_timezone_choices()
    # fallback value
    default = settings.TIME_ZONE


@user_preferences_registry.register
class OcrLanguage(ChoicePreference):
    help_text = 'Language used for OCR processing'
    section = ocr
    name = 'language'
    choices = get_ocr_langs()
    default = get_default_ocr_lang()


@user_preferences_registry.register
class OcrTrigger(ChoicePreference):
    help_text = 'OCR Process Trigger'
    section = ocr
    name = 'trigger'
    choices = [('auto', 'Automatic'), ('manual', 'Manual')]
    default = 'auto'


@user_preferences_registry.register
class LocalizationDate(ChoicePreference):
    help_text = 'Date format'
    section = localization
    name = 'data_format'
    choices = (
        ('%Y-%m-%d', '2020-11-25'),
        ('%a %d %b, %Y', 'Wed 25 Nov, 2020'),
        ('%d %b, %Y', '25 Nov, 2020'),
        ('%m/%d/%Y', '11/25/2020'),
        ('%d/%m/%Y', '25/11/2020'),
        ('%d.%m.%y', '25.11.20'),
        ('%d.%m.%Y', '25.11.2020'),
    )
    default = '%Y-%m-%d'


@user_preferences_registry.register
class LocalizationTime(ChoicePreference):
    help_text = 'Time format'
    section = localization
    name = 'time_format'
    choices = (
        ('%I:%M %p', '9:48 PM'),
        ('%H:%M', '21:48'),
    )
    default = '%H:%M'



import datetime
import logging
import time
import io
import os
import tarfile
import json
from pathlib import PurePath

from django.core.files.temp import NamedTemporaryFile
from mglib.pdfinfo import get_pagecount

from papermerge.core import __version__ as PAPERMERGE_VERSION
from papermerge.core.models import (
    Document,
    User,
    Folder,
    Tag,
    BaseTreeNode
)
from papermerge.core.storage import default_storage
from papermerge.core.utils import remove_backup_filename_id
from papermerge.core.tasks import ocr_document_task

logger = logging.getLogger(__name__)


def backup_documents(
    backup_file: io.BytesIO,
    user: User,
    include_user_password=False
):
    """
    Backup all documents for specific user.

    :include_user_password: if set to True,  will add to backup archive digest
        of user password.
    """

    current_backup = dict()
    current_backup['created'] = datetime.datetime.now()
    current_backup['version'] = PAPERMERGE_VERSION

    with tarfile.open(fileobj=backup_file, mode="w") as backup_archive:

        if user:
            current_backup['documents'] = list()
            _add_user_documents(
                user,
                current_backup,
                backup_archive,
                include_user_in_path=False
            )
        else:
            current_backup['users'] = list()
            for user in User.objects.all():
                current_user = {
                    'username': user.username,
                    'email': user.email,
                    'is_superuser': user.is_superuser,
                    'is_active': user.is_active,
                    'documents': []
                }
                if include_user_password:
                    # raw digest of user's password
                    current_user['password'] = user.password
                _add_user_documents(
                    user,
                    current_user,
                    backup_archive,
                    include_user_in_path=True
                )
                current_backup['users'].append(current_user)

        json_bytes = json.dumps(
            current_backup,
            indent=4,
            default=str
        ).encode('utf-8')

        tarinfo = tarfile.TarInfo(name='backup.json')
        tarinfo.size = len(json_bytes)
        tarinfo.mtime = time.time()

        backup_archive.addfile(tarinfo, io.BytesIO(json_bytes))


def restore_documents(
    restore_file: io.BytesIO,
    user: User,
    skip_ocr=False
):

    restore_file.seek(0)

    with tarfile.open(fileobj=restore_file, mode="r") as restore_archive:
        backup_json = restore_archive.extractfile('backup.json')
        backup_info = json.load(backup_json)

        leading_user_in_path = False
        _user = user
        if not user:
            leading_user_in_path = True
            # user was not specified. It is assument that
            # backup.json contains a list of users.
            # Thus recreate users first.
            for backup_user in backup_info['users']:
                user = User.objects.create(
                    username=backup_user['username'],
                    email=backup_user['email'],
                    is_active=backup_user['is_active'],
                    is_superuser=backup_user['is_superuser']
                )
                # in case --include-user-password switch was used
                # update user (raw digest of) password field
                password = backup_user.get('password')
                if password:
                    user.password = password
                    user.save()

        for restore_file in restore_archive.getnames():

            if restore_file == "backup.json":
                continue

            logger.debug(f"Restoring file {restore_file}...")

            splitted_path = PurePath(restore_file).parts
            base, ext = os.path.splitext(
                remove_backup_filename_id(splitted_path[-1])
            )

            # if there is leading username, remove it.
            if leading_user_in_path:
                username = splitted_path[0]
                _user = User.objects.get(username=username)
                splitted_path = splitted_path[1:]

            if backup_info.get('documents', False):
                backup_info_documents = backup_info['documents']
            else:
                backup_info_documents = _get_json_user_documents_list(
                    backup_info,
                    _user
                )
                leading_user_in_path = True

            for info in backup_info_documents:
                document_info = info
                if info['path'] == restore_file:
                    break

            parent = None

            # variables used only to shorten debug message
            _sp = splitted_path
            _rf = restore_file
            logger.debug(
                f"{_rf}: splitted_path={_sp} len(splitted_path)={len(_sp)}"
            )
            # we first have to create a folder structure
            if len(splitted_path) > 1:
                for folder in splitted_path[:-1]:

                    folder_object = Folder.objects.filter(
                        title=folder,
                        user=_user
                    ).filter(parent=parent).first()

                    if folder_object is None:
                        new_folder = Folder.objects.create(
                            title=folder,
                            parent=parent,
                            user=_user
                        )
                        parent = new_folder
                    else:
                        parent = folder_object

            with NamedTemporaryFile("w+b", suffix=ext) as temp_output:
                logger.debug(f"Extracting {restore_file}...")

                ff = restore_archive.extractfile(restore_file)
                temp_output.write(
                    ff.read()
                )
                temp_output.seek(0)
                size = os.path.getsize(temp_output.name)

                page_count = get_pagecount(temp_output.name)

                if parent:
                    parent_id = parent.id
                else:
                    parent_id = None

                new_doc = Document.objects.create_document(
                    user=_user,
                    title=document_info['title'],
                    size=size,
                    lang=document_info['lang'],
                    file_name=remove_backup_filename_id(splitted_path[-1]),
                    parent_id=parent_id,
                    notes="",
                    page_count=page_count,
                    rebuild_tree=False  # speeds up 100x
                )

                tag_attributes = document_info.get('tags', [])

                for attrs in tag_attributes:
                    attrs['user'] = _user
                    tag, created = Tag.objects.get_or_create(**attrs)
                    new_doc.tags.add(tag)

                default_storage.copy_doc(
                    src=temp_output.name,
                    dst=new_doc.path().url()
                )

            if not skip_ocr:
                ocr_document_task.apply_async(kwargs={
                    'user_id': _user.id,
                    'document_id': new_doc.id,
                    'file_name': new_doc.file_name,
                    'lang': document_info['lang']}
                )


def build_tar_archive(
    fileobj: io.BytesIO,
    node_ids: list
):
    """
    Builds a tar archive with given node ids documents.

    This function is used to download selected documents and folders.
    :fileobj: is a file which will be sent to the client side as
        content disposition.
    """
    with tarfile.open(fileobj=fileobj, mode="w") as archive:
        _rec_tar_archive(
            archive=archive,
            node_ids=node_ids,
            # at the beginning the list of parent names is
            # emtpy
            abspath=[]
        )


def _rec_tar_archive(
    archive,
    node_ids,
    abspath
):
    """
    Recursively builds a tar archive based on folder structure of
    BaseTreeNode instances.

    :node_ids: a list of node ids (node = instance of BaseTreeNode)
    :abspath: a list of folder parent names
    """

    for node in BaseTreeNode.objects.filter(id__in=node_ids):

        if node.is_document():

            arc_path = os.path.join(
                # add document to the current_depth
                *abspath,
                node.title
            )
            archive.add(
                node.absfilepath,
                arcname=arc_path
            )
        elif node.is_folder():

            child_ids = [
                child.id for child in node.get_children()
            ]
            _rec_tar_archive(
                archive,
                child_ids,
                abspath + [node.title]
            )


def _can_restore(restore_file: io.BytesIO):
    with tarfile.open(fileobj=restore_file, mode="r") as restore_archive:
        backup_json = restore_archive.extractfile('backup.json')
        if backup_json is None:
            return False
        current_backup = json.load(backup_json)

        if current_backup.get('version') is not None:
            return True


def _is_valid_user(username: str):
    current_user = User.objects.filter(username=username).first()
    if current_user is not None:
        return True
    else:
        return False


def _createTargetPath(document: Document, include_user_in_path=False):
    """
    Takes a document and traverses the tree to the root noting the path
    :param document: the document you want the path of
    :return: the full path from root to the document including filename
    :rtype str
    """
    # make filename unique
    targetPath = f"{document.file_name}__{document.id}"
    currentNode = document
    while currentNode.parent is not None:
        currentNode = currentNode.parent
        targetPath = os.path.join(
            currentNode.title,
            targetPath
        )

    if include_user_in_path:
        targetPath = os.path.join(
            document.user.username, targetPath
        )

    return targetPath


def _add_current_document_entry(
    document,
    include_user_in_path=False
):
    current_document = {}

    targetPath = _createTargetPath(
        document,
        include_user_in_path=include_user_in_path
    )
    tags = [tag.to_dict() for tag in document.tags.all()]

    current_document['path'] = targetPath
    current_document['lang'] = document.lang
    current_document['title'] = document.title
    current_document['tags'] = tags

    return current_document


def _get_json_user_documents_list(json_backup: dict, user: User):
    for _u in json_backup['users']:
        if _u['username'] == user.username:
            return _u['documents']

    return None


def _add_user_documents(
    user,
    current_backup,
    backup_archive,
    include_user_in_path=False
):
    documents = Document.objects.filter(user=user)

    for current_document_object in documents:  # type: Document
        try:
            backup_archive.add(
                current_document_object.absfilepath,
                arcname=_createTargetPath(
                    current_document_object,
                    include_user_in_path=include_user_in_path
                )
            )
            current_document = _add_current_document_entry(
                current_document_object,
                include_user_in_path=include_user_in_path
            )
            current_backup['documents'].append(
                current_document
            )
        except Exception as e:
            # Log error, but continue backup process
            logger.exception(
                f"Error {e} occurred."
            )



from rest_framework.serializers import Serializer
from rest_framework_json_api import serializers
from rest_framework_json_api.relations import ResourceRelatedField
from rest_framework import serializers as rest_serializers

from papermerge.core.models import (
    BaseTreeNode,
    Folder
)
from papermerge.core.serializers import (
    FolderSerializer,
    DocumentSerializer
)


ONLY_ORIGINAL = 'only_original'
ONLY_LAST = 'only_last'
ZIP = 'zip'
TARGZ = 'targz'


class NodeSerializer(serializers.PolymorphicModelSerializer):
    polymorphic_serializers = [
        FolderSerializer,
        DocumentSerializer
    ]

    parent = ResourceRelatedField(queryset=Folder.objects)

    @classmethod
    def get_polymorphic_serializer_for_instance(cls, instance):
        if instance.is_document:
            return DocumentSerializer

        return FolderSerializer

    def to_representation(self, instance):
        """
        Retrieve the appropriate polymorphic serializer and use this to
        handle representation.
        """
        serializer_class = self.get_polymorphic_serializer_for_instance(
            instance
        )
        inst = instance.document_or_folder
        ret = serializer_class(inst, context=self.context).to_representation(
            inst
        )
        return ret

    class Meta:
        model = BaseTreeNode
        resource_name = 'nodes'
        fields = (
            'title',
            'parent',
            'created_at',
            'updated_at',
        )


class Data_NodeSerializer(Serializer):
    data = NodeSerializer()


class NodeIDSerializer(rest_serializers.Serializer):
    id = rest_serializers.UUIDField()


class NodeMoveSerializer(rest_serializers.Serializer):
    # new parent i.e. target folder
    target_parent = NodeIDSerializer(required=True)
    # nodes to move under the new parent
    nodes = NodeIDSerializer(many=True)


class NodesDownloadSerializer(rest_serializers.Serializer):
    # list of nodes to download
    node_ids = rest_serializers.ListField(
        child=rest_serializers.UUIDField()
    )
    file_name = rest_serializers.CharField(
        max_length=32,
        required=False
    )
    # What to include in downloaded file?
    include_version = rest_serializers.ChoiceField(
        choices=(
            (ONLY_ORIGINAL, 'Only original'),
            (ONLY_LAST, 'Only last'),
        ),
        default=ONLY_LAST
    )
    archive_type = rest_serializers.ChoiceField(
        choices=(
            (TARGZ, '.tar.gz'),
            (ZIP, '.zip')
        ),
        default=ZIP
    )


class InboxCountSerializer(rest_serializers.Serializer):
    count = rest_serializers.IntegerField()


class NodeTagsSerializer(rest_serializers.Serializer):
    # list of tag names
    tags = rest_serializers.ListField(
        child=rest_serializers.CharField()
    )



import uuid
from django.db import models
from django.utils.translation import gettext_lazy as _

from papermerge.core import validators

from taggit.models import TagBase, GenericUUIDTaggedItemBase
from taggit.managers import TaggableManager


class UserTaggableManager(TaggableManager):
    """
    Taggable manager for models with user attribute.

    Model with user attribute means following: that model (say MO) is per user.
    Because tags are per User as well - they (tags) will need to get
    (inherit) user instance from respective model (MO). For this reason,
    save_from_data method is overriden - it passes user attribut to the newly
    saved tag model.

    Example: automate model is per user:

        class Automate(models.Model):
            ...
            tags = UserTaggableManager(
                through=ColoredTag,
                blank=True  # tags are optional
            )
            ...
            user = models.ForeignKey(
                'User',
                models.CASCADE,
                blank=True,
                null=True
            )
            ...
    )
    """

    def save_form_data(self, instance, value):
        rel = getattr(instance, self.name)

        if hasattr(instance, 'user'):
            rel.set(*value, tag_kwargs={'user': instance.user})
        else:
            rel.set(*value)


class Tag(TagBase):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4)

    bg_color = models.CharField(
        _("Background Color"),
        max_length=7,     # RGB color in hex notation
        default='#c41fff',  # purple,
    )

    fg_color = models.CharField(
        _("Foreground Color"),
        max_length=7,      # RGB color in hex notation
        default='#FFFFFF'  # white
    )

    description = models.TextField(
        _("Description (optional)"),
        max_length=1024,
        blank=True,
        null=True
    )

    # a pinned tag may be displayed for example under "Documents" menu
    # of left side bar. It serves as shortcut for user to quickly filter
    # folders/documents with that particular tag.
    pinned = models.BooleanField(
        default=False,
        help_text=_(
            "Pinned tag will be displayed under Documents menu. "
            "It serves as shortcut to quickly filter "
            "folders/documents associated with this tag"
        )
    )

    # each user has his/her set of tags
    user = models.ForeignKey('User', models.CASCADE)
    name = models.CharField(
        verbose_name=_("name"),
        unique=False,
        max_length=100,
        validators=[validators.safe_character_validator]
    )

    class Meta:
        verbose_name = _("Tag")
        verbose_name_plural = _("Tags")
        ordering = ['name']
        unique_together = ['name', 'user']


class ColoredTag(GenericUUIDTaggedItemBase):
    tag = models.ForeignKey(
        Tag,
        on_delete=models.CASCADE,
        related_name="%(app_label)s_%(class)s_items",
    )



import re
import logging

from channels.middleware import BaseMiddleware
from knox.models import AuthToken
from knox.settings import CONSTANTS
from knox.crypto import hash_token

from channels.db import database_sync_to_async

logger = logging.getLogger(__name__)

try:
    from hmac import compare_digest
except ImportError:
    def compare_digest(a, b):
        return a == b


HEADER_NAME = 'authorization'
TOKEN_NAME = 'token'
SEC_WEBSOCKET_PROTOCOL = 'sec-websocket-protocol'
ACCESS_TOKEN_NAME = 'access_token'


@database_sync_to_async
def get_user(token_key):
    try:
        auth_token = AuthToken.objects.get(
            token_key=token_key[:CONSTANTS.TOKEN_KEY_LENGTH]
        )
    except AuthToken.DoesNotExist:
        return None

    digest = hash_token(token_key)

    if compare_digest(digest, auth_token.digest):
        if not auth_token.user.is_active:
            return None

        return auth_token.user

    return None


def extract_from_auth_header(value: str) -> str:

    if not value:
        return None

    try:
        token_identifier, token_value = re.split(r'\s+', value.strip())
        if token_identifier:
            token_identifier = token_identifier.strip().lower()
        if token_identifier == TOKEN_NAME:
            return token_value.strip()
    except ValueError:
        logger.warning(
            f"Poorly formatted Authorization header '{value}'"
        )
        return None


def extract_from_sec_websocket_protocol_header(value: str) -> str:
    if not value:
        return None

    try:
        token_access_identifier, token_value = value.split(',')
        if token_access_identifier:
            token_access_identifier = token_access_identifier.strip().lower()
        if token_access_identifier == ACCESS_TOKEN_NAME:
            return token_value.strip()
    except ValueError:
        logger.warning(
            f"Poorly formatted Sec-WebSocket-Protocol header {value}"
        )
        return None


def extract_token(headers):
    """
    Returns token from list of headers

    Token can be passed in two ways (or formats):
    1. Authorization: Token <value>
    2. Sec-WebSocket-Protocol: access_token, <value>

    First case may be used by command line utilities.
    Second case is used by browser clients via

        WebSocket(url, ['access_token', token])

    If no token header was found, returns None.
    """
    for _key, _value in headers:
        key = _key.decode().lower()
        value = _value.decode()
        if key == HEADER_NAME:
            return extract_from_auth_header(value)
        if key == SEC_WEBSOCKET_PROTOCOL:
            return extract_from_sec_websocket_protocol_header(value)

    return None


class TokenAuthMiddleware(BaseMiddleware):

    async def __call__(self, scope, receive, send):
        token_key = extract_token(headers=scope['headers'])

        if token_key is None:
            scope['user'] = None
        else:
            scope['user'] = await get_user(token_key)

        return await super().__call__(scope, receive, send)



import os
import zipfile
import tarfile

from django.core.files.temp import NamedTemporaryFile

from papermerge.core.models import Document, BaseTreeNode
from papermerge.core.serializers.node import (
    ONLY_ORIGINAL,
    ONLY_LAST,
    ZIP
)


class NodesDownload:
    """
    Base class for different types of nodes download:
        * document file (e.g. invoice.pdf)
        * zip archive (e.g. my_invoices.zip)
        * targz archive (e.g. my_invoices.tar.gz
    """
    def __init__(
        self,
        node_ids,
        file_name=None,
        include_version=ONLY_LAST
    ):
        self._file_name = file_name
        self._include_version = include_version
        self._node_ids = node_ids

    def _create(self):
        raise Exception("Not Implemented")

    def get_content(self):
        file_handle = self._create()
        file_handle.seek(0)
        data = file_handle.read()
        file_handle.close()

        return data

    def wants_only_orignal(self):
        return self._include_version == ONLY_ORIGINAL

    def wants_only_last(self):
        return self._include_version == ONLY_LAST

    def _recursive_create_archive(self, archive, node_ids, abspath):
        for node in BaseTreeNode.objects.filter(id__in=node_ids):
            if node.is_document():
                arcname = os.path.join(*abspath, node.idified_title)
                if self.wants_only_last():
                    doc_version = node.versions.last()
                else:
                    doc_version = node.versions.first()

                self.archive_add(
                    archive=archive,
                    abs_file_path=doc_version.abs_file_path(),
                    arcname=arcname
                )
            else:
                child_ids = [child.id for child in node.get_children()]
                self._recursive_create_archive(
                    archive,
                    child_ids,
                    abspath + [node.idified_title]
                )

    def archive_add(self, abs_file_path, arcname):
        raise Exception("Not Implemented")

    @property
    def file_name(self):
        raise Exception("Not Implemented")

    @property
    def content_type(self):
        raise Exception("Not Implemented")

    @property
    def content_disposition(self):
        return f"attachment; filename={self.file_name}"


class NodesDownloadZip(NodesDownload):
    """
    Creates zip archive from a list of nodes. It preserves folder structure.

    Usage:

    nodes_download = NodesDownloadZip(
        node_ids=[1, 2, 3],
        file_name="archive.zip",
        include_version="only_last"
    )

    # returns zip archives content (as bytes)
    data = nodes_download.get_content()
    """
    def _create(self):
        temp_file_obj = NamedTemporaryFile(prefix='download_')
        archive = zipfile.ZipFile(temp_file_obj, mode='w')
        self._recursive_create_archive(
            archive=archive,
            node_ids=self._node_ids,
            abspath=[]
        )
        temp_file_obj.seek(0)
        return temp_file_obj

    def archive_add(self, archive, abs_file_path, arcname):
        archive.write(abs_file_path, arcname=arcname)

    @property
    def file_name(self):
        if self._file_name:
            return self._file_name

        return "archive.zip"

    @property
    def content_type(self):
        return "application/zip"

    def __str__(self):
        return f'NodesDownloadZip(node_ids={self._node_ids})'


class NodesDownloadTarGz(NodesDownload):
    """
    Creates targz archive from a list of nodes. It preserves folder structure.

    Usage:

    nodes_download = NodesDownloadTarGz(
        node_ids=[1, 2, 3],
        file_name="archive.tar.gz",
        include_version="only_last"
    )

    # returns tar.gz archives content (as bytes)
    data = nodes_download.get_content()
    """
    def _create(self):
        temp_file_obj = NamedTemporaryFile(prefix='download_')
        archive = tarfile.open(temp_file_obj.name, mode='w:gz')
        self._recursive_create_archive(
            archive=archive,
            node_ids=self._node_ids,
            abspath=[]
        )
        temp_file_obj.seek(0)
        return temp_file_obj

    def archive_add(self, archive, abs_file_path, arcname):
        archive.add(abs_file_path, arcname=arcname)

    @property
    def file_name(self):
        if self._file_name:
            return self._file_name

        return "archive.tar.gz"

    @property
    def content_type(self):
        return "application/x-gtar"

    def __str__(self):
        return f'NodesDownloadTarGz(node_ids={self._node_ids})'


class NodesDownloadDocument(NodesDownload):
    """
    Returns document file data of the document in node_ids[0].

    Note that node_ids must have only one item in the list.

    Usage:

    nodes_download = NodesDownloadDocument(
        node_ids=[1],
        include_version="only_last"
    )

    # returns document's last version file content (as bytes)
    data = nodes_download.get_content()
    """
    def _create(self):
        abs_path = self.get_document_file_abs_path()
        return open(abs_path, 'rb')

    def get_document_version(self):
        doc = Document.objects.get(pk=self._node_ids[0])
        if self.wants_only_last():
            return doc.versions.last()

        return doc.versions.first()

    def get_document_file_abs_path(self):
        """Returns the absolute path to the document file to be downloaded"""
        doc_version = self.get_document_version()
        abs_file_path = doc_version.abs_file_path()

        return abs_file_path

    @property
    def file_name(self):
        if self._file_name:
            return self._file_name

        doc_version = self.get_document_version()
        return doc_version.file_name

    @property
    def content_type(self):
        # TODO: add correct content type depending on the
        # document file
        return "application/pdf"

    def __str__(self):
        return f'NodesDownloadDocument(node_ids={self._node_ids})'

    def __repr__(self):
        return str(self)


def is_single_document_node(node_ids, include_version):
    """
    Returns True only if all following conditions are True:
        * node_ids has only one item
        * include_version is either 'only_last' or 'only_original'
        * Document model referenced by ID=node_ids[0] exists

    Returns False if one of above conditions is False.

    :param node_ids: list of node IDs
    :param include_version: a list of strings
    :return:
    """
    if len(node_ids) == 1 and include_version in (ONLY_LAST, ONLY_ORIGINAL):
        return Document.objects.filter(pk=node_ids[0]).exists()

    return False


def get_nodes_download(
    node_ids,
    file_name=None,
    include_version=ONLY_LAST,
    archive_type=ZIP
):
    """
    Helper function which returns correct instance/version
    of NodesDownload class

    Arguments:

    * `node_ids` a list of node IDs to download
    * `file_name` file name of end archive/document file
    * `include_version` can be either 'only_last' or 'only_original'
    * `archive_type` can be either 'zip' or 'targz'

    Returns:
        NodesDownloadDocument or NodesDownloadZip or NodesDownloadTarGzFile
        instance depending on the content of `node_ids` and `include_version`
        arguments
    """
    if is_single_document_node(node_ids, include_version):
        return NodesDownloadDocument(
            node_ids=node_ids,
            file_name=file_name,
            include_version=include_version
        )

    if archive_type == ZIP:
        return NodesDownloadZip(
            node_ids,
            file_name,
            include_version=include_version
        )

    return NodesDownloadTarGz(
        node_ids,
        file_name,
        include_version=include_version
    )



from django.apps import AppConfig


class NotificationsConfig(AppConfig):
    name = 'papermerge.notifications'
    label = 'notifications'

    def ready(self):
        from papermerge.core import signals  # noqa



import io
import os
import logging

from django.utils.translation import gettext_lazy as _

from celery import shared_task
from papermerge.core.ocr.document import ocr_document
from papermerge.core.storage import abs_path, get_storage_instance

from .models import (
    Document,
    DocumentVersion,
    Folder,
    Page
)

logger = logging.getLogger(__name__)


@shared_task
def delete_user_data(user_id):
    """Delete ALL associated user data (invoked when user is deleted)"""
    logger.debug(f'Deleting user {user_id} storage data')
    storage = get_storage_instance()
    storage.delete_user_data(user_id=user_id)


@shared_task(acks_late=True, reject_on_worker_lost=True)
def ocr_document_task(
    document_id,
    lang,
    user_id,  # UUID of the user who initiated OCR of the document
    namespace=None
):
    """
    OCRs the document.

    On success returns ``document_id``
    If something went wrong returns ``None``.

    Returning ``document_id`` on success crucial, as ``ocr_document_task``
    is chained with other celery tasks which will receive as
    first argument returned value of this task (i.e. ``document_id``).

    If you use `acks_late` then the worker will remove the item from the queue
    at the end of the task rather than the beginning.
    However, if the worker process is killed the task is still acknowledged even
    if it wasn't completed. `reject_on_worker_lost` will re-queue the message
    if the above event happens so you won't lose the task.
    """
    doc = Document.objects.get(pk=document_id)
    user_id = doc.user.id
    doc_version = doc.versions.last()

    logger.debug(
        'ocr_document_task: ocr start'
        f'document_id={document_id} namespace={namespace} '
        f'lang={lang}'
    )

    ocr_document(
        user_id=user_id,
        document_id=document_id,
        file_name=doc_version.file_name,
        lang=lang,
        namespace=namespace,
        version=doc_version.number,
        target_version=doc_version.number + 1
    )

    logger.debug(
        'ocr_document_task: ocr end'
        f'document_id={document_id} namespace={namespace} '
        f'lang={lang}'
    )

    logger.debug(
        'ocr_document_task: successfully complete'
        f'document_id={document_id} namespace={namespace} '
        f'lang={lang}'
    )

    return document_id


@shared_task
def post_ocr_document_task(document_id, namespace=None):
    """
    Task to run immediately after document OCR is complete

    This task guarantees that `increment_document_version` will run
    before `update_document_pages`.
    """
    logger.debug(f'post_ocr_task_task doc_id={document_id}')

    increment_document_version(document_id, namespace)
    update_document_pages(document_id, namespace)

    # generate previews for newly created document version (which has OCR)
    doc = Document.objects.get(pk=document_id)
    doc_version = doc.versions.last()

    generate_page_previews_task.delay(str(doc_version.id))

    return document_id


@shared_task
def generate_page_previews_task(document_version_id):
    document_version = DocumentVersion.objects.get(id=document_version_id)

    doc = document_version.document
    logger.debug(f"Generating previews doc_id={doc.id}")

    document_version.generate_previews()

    return document_version_id


def increment_document_version(document_id, namespace=None):
    logger.debug(
        'increment_document_version: '
        f'document_id={document_id} namespace={namespace}'
    )

    doc = Document.objects.get(pk=document_id)
    lang = doc.lang
    doc_version = doc.versions.last()

    new_doc_version = DocumentVersion(
        document=doc,
        number=doc_version.number + 1,
        file_name=doc_version.file_name,
        size=0,  # TODO: set to newly created file size
        page_count=doc_version.page_count,
        lang=lang,
        short_description=_("with OCRed text layer")
    )
    new_doc_version.save()

    logger.debug(
        'ocr_document_task: creating pages'
        f'document_id={document_id} namespace={namespace} '
        f'lang={lang}'
    )

    for page_number in range(1, new_doc_version.page_count + 1):
        Page.objects.create(
            document_version=new_doc_version,
            number=page_number,
            page_count=new_doc_version.page_count,
            lang=lang
        )


def update_document_pages(document_id, namespace=None):
    """
    Updates document latest versions's ``text`` field

    ``text`` field is updated on the last document version instance
    as well as on each of last document versions' page.

    In case when a particular file with ``page.txt_url`` does not exist,
    page content (to the precise, the lack of page content) will
    be replaced with empty string (io.String('')).

    In particular when no OCR was performed yet each individual
    page as well as document versions's ``text`` fields will be
    updated with empty strings.
    """

    logger.debug(
        'update_document_pages: '
        f'document_id={document_id} namespace={namespace}'
    )

    doc = Document.objects.get(pk=document_id)
    doc_version = doc.versions.last()
    streams = []

    for page in doc_version.pages.order_by('number'):
        url = abs_path(page.txt_url)
        if os.path.exists(url):
            streams.append(open(url))
        else:
            streams.append(io.StringIO(''))

    doc_version.update_text_field(streams)


def norm_pages_from_doc(document):
    logger.debug(f"Normalizing document {document.id}")
    for page in Page.objects.filter(document=document):
        page.norm()


def norm_pages_from_folder(folder):
    for descendent in folder.get_descendants():
        if isinstance(descendent, Document):
            norm_pages_from_doc(descendent)
        elif isinstance(descendent, Folder):
            norm_pages_from_folder(descendent)
        else:
            raise ValueError("Unexpected value for descendent instance")


@shared_task
def normalize_pages(origin):
    """
    Normalize Pages. The normalization was triggered model origin.
    origin can be either a Folder or a Document
    """
    if isinstance(origin, Document):
        norm_pages_from_doc(origin)
    elif isinstance(origin, Folder):
        norm_pages_from_folder(origin)



import logging
import os
import shutil
from os import listdir
from os.path import isdir, join

from .path import DocumentPath, PagePath, AUX_DIR_SIDECARS, AUX_DIR_DOCS
from .utils import safe_to_delete

logger = logging.getLogger(__name__)


class Storage:
    """
    Default Storage class which works with DocumentPath and PagePath
    on local host filesystem
    """

    def __init__(self, location=None, **kwargs):
        # by default, this will be something like
        # settings.MEDIA_ROOT
        self._location = location

    @property
    def location(self):
        return self._location

    def upload(self, doc_path_url, **kwargs):
        pass

    def download(self, doc_path_url, **kwargs):
        pass

    def delete_user_data(self, user_id: str):
        folder1_to_delete = os.path.join(
            self.abspath(AUX_DIR_DOCS),
            f'user_{user_id}'
        )
        folder2_to_delete = os.path.join(
            self.abspath(AUX_DIR_SIDECARS),
            f'user_{user_id}'
        )

        self.safe_delete_folder(folder1_to_delete)
        self.safe_delete_folder(folder2_to_delete)

    def safe_delete_folder(self, abs_path_to_folder_to_delete: str):
        logger.debug(
            f'Safely deleting content of {abs_path_to_folder_to_delete}'
        )
        if safe_to_delete(
            abs_path_to_folder_to_delete
        ):
            shutil.rmtree(abs_path_to_folder_to_delete)
            if os.path.exists(abs_path_to_folder_to_delete):
                os.rmdir(abs_path_to_folder_to_delete)

    def make_sure_path_exists(self, filepath):
        logger.debug(f"make_sure_path_exists {filepath}")
        dirname = os.path.dirname(filepath)
        os.makedirs(
            dirname,
            exist_ok=True
        )

    def get_versions(self, doc_path):
        """
        Returns a list of (all) ordered versions
        of specific doc_path. Versions
        start with 0. Examples of return values:

        - [0, 1, 2, 3] = 4 versions of the document
        - [ 0 ] = only one version (original)

        To count versions it just counts number of subfolders
        in specific document folder. Versions are
        stored in subfolders named v1, v2, v3, ...
        """
        abs_dirname_docs = self.path(
            doc_path.dirname_docs
        )
        try:
            only_dirs = [
                fi for fi in listdir(abs_dirname_docs) if isdir(
                    join(abs_dirname_docs, fi)
                )
            ]
        except FileNotFoundError:
            # in tests, document folders are not always created.
            # If no document folder is found, just return [ 0 ]
            # i.e that document has only one single version and it
            # is the latest one.
            return [0]

        dirs_count = len(only_dirs)

        return list(range(0, dirs_count + 1))

    def get_pagecount(self, doc_path):
        """
        Returns total number of pages for this doc_path.
        Total number of pages = number of page_xy.txt files
        in pages_dirname folder.
        """
        doc_path_pointing_to_results = DocumentPath.copy_from(
            doc_path, aux_dir="results"
        )
        pages_dir = self.abspath(doc_path_pointing_to_results.pages_dirname())

        only_dirs = [
            fi for fi in listdir(pages_dir) if isdir(join(pages_dir, fi))
        ]
        return len(only_dirs)

    def abspath(self, _path):
        if isinstance(_path, DocumentPath):
            return os.path.join(
                self.location, _path.url
            )
        elif isinstance(_path, PagePath):
            return os.path.join(
                self.location, _path.url
            )

        return os.path.join(
            self.location, _path
        )

    def path(self, _path):
        return self.abspath(_path)

    def delete_doc(self, doc_path: DocumentPath):
        """
        Receives a DocumentPath instance
        """
        # where original documents and their versions are stored
        abs_dirname_docs = self.path(
            doc_path.dirname_docs
        )
        # where OCRed information and generated thumbnails
        # are stored
        abs_dirname_sidecars = self.path(
            doc_path.dir_sidecars
        )
        # Before recursively deleting everything in folder
        # double check that there are only
        # .pdf, .txt, .hocr, .jpg files.
        if safe_to_delete(
            abs_dirname_docs
        ):
            shutil.rmtree(abs_dirname_docs)
            if os.path.exists(abs_dirname_docs):
                os.rmdir(abs_dirname_docs)

        if safe_to_delete(
            abs_dirname_sidecars
        ):
            shutil.rmtree(abs_dirname_sidecars)
            if os.path.exists(abs_dirname_sidecars):
                os.rmdir(abs_dirname_sidecars)

    def copy_doc(self, src: DocumentPath, dst: DocumentPath):
        """
        copy given file src file path to destination
        as absolute doc_path
        """

        dirname = os.path.dirname(
            self.abspath(dst)
        )
        if not os.path.exists(
            dirname
        ):
            os.makedirs(
                dirname, exist_ok=True
            )
        logger.debug(
            f"copy_doc: {src} to {dst}"
        )
        shutil.copyfile(
            self.abspath(src),
            self.abspath(dst)
        )

    def exists(self, _path):
        return os.path.exists(
            self.path(_path)
        )

    def copy_page_txt(self, src: PagePath, dst: PagePath):
        logger.debug(f"copy_page_txt src={src.txt_url} dst={dst.txt_url}")
        src_txt = self.abspath(src.txt_url)
        dst_txt = self.abspath(dst.txt_url)

        self.make_sure_path_exists(
            self.abspath(dst.txt_url)
        )

        shutil.copy(src_txt, dst_txt)

    def copy_page_jpg(self, src: PagePath, dst: PagePath):
        logger.debug(f"copy_page_jpg src={src.jpg_url} dst={dst.jpg_url}")
        src_jpg = self.abspath(src.jpg_url)
        dst_jpg = self.abspath(dst.jpg_url)

        self.make_sure_path_exists(
            self.abspath(dst.jpg_url)
        )
        shutil.copy(src_jpg, dst_jpg)

    def copy_page_hocr(self, src: PagePath, dst: PagePath):
        logger.debug(f"copy_page_hocr: src={src.hocr_url} dst={dst.hocr_url}")
        src_hocr = self.abspath(src.hocr_url)
        dst_hocr = self.abspath(dst.hocr_url)

        self.make_sure_path_exists(
            self.abspath(dst.hocr_url)
        )

        shutil.copy(src_hocr, dst_hocr)

    def copy_page_svg(self, src: PagePath, dst: PagePath):
        logger.debug(f"copy_page_svg: src={src.svg_url} dst={dst.svg_url}")
        src_svg = self.abspath(src.svg_url)
        dst_svg = self.abspath(dst.svg_url)

        self.make_sure_path_exists(self.abspath(dst.svg_url))

        shutil.copy(src_svg, dst_svg)

    def copy_page_preview(self, src: PagePath, dst: PagePath):
        logger.debug(
            f"copy_page_preview: src={src.preview_url} dst={dst.preview_url}"
        )
        src_preview = self.abspath(src.preview_url)
        dst_preview = self.abspath(dst.preview_url)

        self.make_sure_path_exists(self.abspath(dst.preview_url))

        shutil.copy(src_preview, dst_preview)

    def copy_page(self, src: PagePath, dst: PagePath):
        """
        Copies page data from source folder/path to page destination folder/path

        Page data are files with 'txt', 'hocr', 'jpg', 'svg' extentions.
        """
        for inst in [src, dst]:
            if not isinstance(inst, PagePath):
                raise ValueError("copy_page accepts only PagePath instances")

        # copy .txt file
        if self.exists(src.txt_url):
            self.copy_page_txt(src=src, dst=dst)
        else:
            logger.debug(f"txt does not exits {src.txt_url}")

        # hocr
        if self.exists(src.hocr_url):
            self.copy_page_hocr(src=src, dst=dst)
        else:
            logger.debug(f"hocr does not exits {src.hocr_url}")

        if self.exists(src.jpg_url):
            self.copy_page_jpg(src=src, dst=dst)
        else:
            logger.debug(f"jpg does not exits {src.jpg_url}")

        if self.exists(src.svg_url):
            self.copy_page_svg(src=src, dst=dst)
        else:
            logger.debug(f"svg does not exits {src.svg_url}")

        if self.exists(src.preview_url):
            self.copy_page_preview(src=src, dst=dst)
        else:
            logger.debug(f"preview does not exits {src.preview_url}")

    def reorder_pages(self, doc_path, new_order):
        """
        Reorders pages in the document pointed by doc_path.
        doc_path is an instance of mglib.path.DocumentPath

        In case of success returns document's new version.

        new_order is a list of following format:

            [
                {'page_num': 2, page_order: 1},
                {'page_num': 1, page_order: 2},
                {'page_num': 3, page_order: 3},
                {'page_num': 4, page_order: 4},
            ]
        Example above means that in current document of 4 pages,
        first page was swapped with second one.
        page_num    = older page order
        page_order  = current page order
        So in human language, each hash is read:
            <page_num> now should be <page_order>
        """
        src_doc_path = doc_path
        dst_doc_path = DocumentPath.copy_from(
            src_doc_path,
            version=doc_path.version + 1
        )
        self.make_sure_path_exists(
            self.abspath(dst_doc_path)
        )

        # replace stapler!
        # stapler.reorder_pages(
        #    src=self.abspath(src_doc_path),
        #    dst=self.abspath(dst_doc_path),
        #    new_order=new_order
        # )

        page_count = self.get_pagecount(doc_path)

        if len(new_order) > page_count:
            logger.error(
                f"deleted_pages({new_order}) > page_count({page_count})"
            )
            return

        # steps were removed
        # for item in new_order:
        #    for step in Steps():
        #        src_page_path = PagePath(
        #            document_path=src_doc_path,
        #            page_num=int(item['page_num']),
        #            step=step,
        #            page_count=len(new_order)
        #        )
        #        dst_page_path = PagePath(
        #            document_path=dst_doc_path,
        #            page_num=int(item['page_order']),
        #            step=step,
        #            page_count=len(new_order)
        #        )
        #        self.copy_page(
        #            src_page_path=src_page_path,
        #            dst_page_path=dst_page_path
        #        )

        return doc_path.version + 1

    def delete_pages(
        self,
        doc_path,
        page_numbers,
        skip_migration=False
    ):
        """
        Delets pages in the document pointed by doc_path.
        doc_path is an instance of mglib.path.DocumentPath

        In case of success returns document's new version.
        """

        if not isinstance(page_numbers, list):
            logger.error("Expecting list argument")
            return False

        src_doc_path = doc_path
        dst_doc_path = DocumentPath.copy_from(
            src_doc_path,
            version=doc_path.version + 1
        )
        self.make_sure_path_exists(
            self.abspath(dst_doc_path)
        )

        if skip_migration:
            return doc_path.version + 1

        page_count = self.get_pagecount(doc_path)

        if len(page_numbers) > page_count:
            logger.error(
                f"deleted_pages({page_numbers}) > page_count({page_count})"
            )
            return

        return doc_path.version + 1

    def paste_pages(
        self,
        dest_doc_path,
        data_list,
        dest_doc_is_new=False,
        after_page_number=False,
        before_page_number=False
    ):
        """
        Pastes pages in the document pointed by dest_doc_path
        from src_doc_path. Both dest and src are instances of
        mglib.path.DocumentPath
        """
        next_version = 0
        if dest_doc_is_new:
            # document is new, start version with 0
            next_version = 0
        else:
            # destination document is not new, increment its version
            next_version = dest_doc_path.version + 1

        next_ver_dp = DocumentPath.copy_from(
            dest_doc_path,
            version=next_version
        )
        self.make_sure_path_exists(
            self.abspath(next_ver_dp)
        )

        # stapler.paste_pages(
        #    src=self.abspath(dest_doc_path),
        #    dst=self.abspath(next_ver_dp),
        #    data_list=data_list,
        #    dst_doc_is_new=dest_doc_is_new,
        #    after_page_number=after_page_number,
        #    before_page_number=before_page_number
        # )

        if not dest_doc_is_new:
            # migrate document's own pages from previous
            # version (this differs from pasting into newly
            # created docs)
            pcount = self.get_pagecount(dest_doc_path)
            data_list.insert(
                0,
                {
                    'doc_path': dest_doc_path,
                    'page_nums': list(range(1, pcount + 1))
                }
            )

        # dest_page_num = 1
        # dest_page_count = sum([
        #    len(item['page_nums']) for item in data_list
        # ])
        # for item in data_list:
        #    src_path = item['doc_path']
        #    for page_num in item['page_nums']:
        #        for step in Steps():
        #            src_page_path = PagePath(
        #                document_path=src_path,
        #                page_num=int(page_num),
        #                step=step,
        #                page_count=self.get_pagecount(src_path)
        #            )
        #            dst_page_path = PagePath(
        #                document_path=next_ver_dp,
        #                page_num=dest_page_num,
        #                step=step,
        #                page_count=dest_page_count
        #            )
        #            logger.debug(f"src={src_page_path}  dst={dst_page_path}")
        #            self.copy_page(
        #                src_page_path=src_page_path,
        #                dst_page_path=dst_page_path
        #            )
        #        dest_page_num += 1

        return next_version


class FileSystemStorage(Storage):
    pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['conf',
 'core',
 'core.auth',
 'core.lib',
 'core.management',
 'core.management.commands',
 'core.middleware',
 'core.migrations',
 'core.models',
 'core.ocr',
 'core.openapi',
 'core.serializers',
 'core.views',
 'notifications',
 'notifications.consumers',
 'search',
 'search.migrations']

package_data = \
{'': ['*'],
 'core': ['locale/de/LC_MESSAGES/*', 'locale/fr/LC_MESSAGES/*'],
 'search': ['templates/search/indexes/core/*']}

install_requires = \
['Whoosh>=2.7.4,<3.0.0',
 'celery[gevent,redis]>=5.2,<6.0',
 'channels-redis>=3.2,<4.0',
 'channels>=3.0,<4.0',
 'configula>=0.5,<0.6',
 'daphne>=3.0.2,<4.0.0',
 'django-celery-results>=2.2,<3.0',
 'django-cors-headers>=3.9.0,<4.0.0',
 'django-dynamic-preferences>=1.13,<2.0',
 'django-haystack>=3.2.1,<4.0.0',
 'django-modelcluster>=6.0,<7.0',
 'django-rest-knox>=4.2.0,<5.0.0',
 'django-taggit>=3.0.0,<4.0.0',
 'django>=4.0,<5.0',
 'django_filter>=21.1,<22.0',
 'djangorestframework-jsonapi>=5.0.0,<6.0.0',
 'djangorestframework>=3.13.1,<4.0.0',
 'drf-spectacular-sidecar>=2022.3.21,<2023.0.0',
 'drf-spectacular>=0.22.0,<0.23.0',
 'elasticsearch>=7,<8',
 'lxml>=4.9.0,<5.0.0',
 'ocrmypdf-papermerge>=0.4.5,<0.5.0',
 'ocrmypdf>=13.5.0,<14.0.0',
 'pdf2image>=1.16.0,<2.0.0',
 'persisting-theory>=1.0,<2.0',
 'psycopg2-binary>=2.9.2,<3.0.0',
 'psycopg2>=2.9.2,<3.0.0',
 'python-magic>=0.4,<0.5',
 'pyyaml>=6.0,<7.0',
 'redis>=4.3.3,<5.0.0',
 'uWSGI>=2.0.20,<3.0.0',
 'uritemplate>=4.1.1,<5.0.0',
 'yapian-haystack>=3.1.0,<4.0.0']

setup_kwargs = {
    'name': 'papermerge-core',
    'version': '2.1.3',
    'description': 'Open source document management system for digital archives',
    'long_description': "[![Tests](https://github.com/papermerge/papermerge-core/actions/workflows/tests.yml/badge.svg)](https://github.com/papermerge/papermerge-core/actions/workflows/tests.yml)\n\n# Papermerge REST API Server\n\nThis python package is the heart of Papermerge project. It consists of a set\nof reusable Django apps which are consumed across different bundles of\nPapermerge Document Management System (DMS).\n\nTechnically speaking, it contains following Django apps:\n\n* ``papermerge.core`` - the epicenter of Papermerge DMS project\n* ``papermerge.notifications`` - Django Channels app for sending notifications via websockets\n* ``papermerge.search`` - RESTful search. Supports four backends: [Xapian](https://getting-started-with-xapian.readthedocs.io/en/latest/),\n  [Whoosh](https://whoosh.readthedocs.io/en/latest/intro.html), [Elasticsearch](https://github.com/elastic/elasticsearch),\n  [Solr](https://solr.apache.org/).\n\n\n## What is Papermerge?\n\nPapermerge is an open source document management system (DMS) primarily\ndesigned for archiving and retrieving your digital documents. Instead of\nhaving piles of paper documents all over your desk, office or drawers - you\ncan quickly scan them and configure your scanner to directly upload to\nPapermerge DMS. Papermerge DMS on its turn will extract text data from the\nscanned documents using Optical Character Recognition (OCR) technology the\nindex it and make it searchable. You will be able to quickly find any\n(scanned!) document using full text search capabilities.\n\nPapermerge is perfect tool to manage documents in PDF, JPEG, TIFF and PNG formats.\n\n## Features Highlights\n\n* OpenAPI compliant REST API\n* Works well with PDF documents\n* OCR (Optical Character Recognition) of the documents (uses [OCRmyPDF](https://github.com/ocrmypdf/OCRmyPDF))\n* Full Text Search of the scanned documents (supports four search engine backends, uses [Xapian](https://getting-started-with-xapian.readthedocs.io/en/latest/) by default)\n* Document Versions\n* Tags - assign colored tags to documents or folders\n* Documents and Folders - users can organize documents in folders\n* Multi-User (supports user groups)\n* User permissions management\n* Page Management - delete, reorder, cut & paste pages (uses [PikePDF](https://github.com/pikepdf/pikepdf))\n\n\n## Documentation\n\nFor an overview on REST API is available [here](https://docs.papermerge.io/REST%20API/index.html).\n\nDetailed online REST API reference can be viewed as:\n\n- [redoc](https://docs.papermerge.io/redoc/)\n- [swagger](https://docs.papermerge.io/swagger-ui/)\n\nNote that REST API reference documentation is generated from\nOpenAPI schema. OpenAPI schema is stored in its own dedicated\nrepository [papermerge/openapi-schema](https://github.com/papermerge/openapi-schema).\n\nPapermerge DMS documentation is available at [https://docs.papermerge.io](https://docs.papermerge.io/)\n\n## Docker\n\nIn order to start Papermerge REST API server as docker image use following command:\n\n    docker run -p 8000:8000 \\\n        -e PAPERMERGE__MAIN__SECRET_KEY=abc \\\n        -e DJANGO_SUPERUSER_PASSWORD=123 \\\n        papermerge/papermerge:latest\n\n\nIf you want initial superuser to have another username (e.g. john), use\n`DJANGO_SUPERUSER_USERNAME` environment variable:\n\n    docker run -p 8000:8000 \\\n        -e PAPERMERGE__MAIN__SECRET_KEY=abc \\\n        -e DJANGO_SUPERUSER_PASSWORD=123 \\\n        -e DJANGO_SUPERUSER_USERNAME=john \\\n        papermerge/papermerge:latest\n\nFor full list of supported environment variables check [online documentation](https://docs.papermerge.io/Settings/index.html).\n\n## Docker Compose\n\nBy default Papermerge REST API server uses sqlite3 database. In order to use PostgreSQL use following docker compose file:\n\n    version: '3.7'\n    services:\n      app:\n        image: papermerge/papermerge\n        environment:\n          - PAPERMERGE__MAIN__SECRET_KEY=abc\n          - DJANGO_SUPERUSER_PASSWORD=12345\n          - PAPERMERGE__DATABASE__TYPE=postgres\n          - PAPERMERGE__DATABASE__USER=postgres\n          - PAPERMERGE__DATABASE__PASSWORD=123\n          - PAPERMERGE__DATABASE__NAME=postgres\n          - PAPERMERGE__DATABASE__HOST=db\n        ports:\n          - 8000:8000\n        depends_on:\n          - db\n      db:\n        image: bitnami/postgresql:14.4.0\n        volumes:\n          - postgres_data:/var/lib/postgresql/data/\n        environment:\n          - POSTGRES_PASSWORD=123\n    volumes:\n        postgres_data:version: '3.7'\n\nAbove mentioned docker compose file can be used to start Papermerge REST API server which will use PostgreSQL database to store data.\n\nFor detailed description on how to start Papermerge DMS using docker compose read\n[Docker Compose/Detailed Explanation](https://docs.papermerge.io/Installation/docker-compose.html#detailed-explanation)\nsection in online docs.\n\n## Tests\n\nTest suite is divided into two big groups:\n\n1. tests.core\n2. tests.search\n\n\nFirst group is concerned with tests which do not depend on elasticsearch while\nsecond one **tests.search** is concerned with tests for which **depend on elasticsearch**\nand as result run very slow (hence the grouping). In\norder to run `tests.core` tests you need to have redis up and running; in\norder to run `test.search` you need to both **redis and elasticsearch** up and\nrunning.\n\nBefore running core tests suite, make sure redis service is up and running. Run tests:\n\n     poetry run task test-core\n\nBefore running search tests suite, make sure both **redis and elasticsearch**\nservices are up and running:\n\n     poetry run task test-search\n\nIn order to run all tests suite (core + search):\n\n    poetry run task test\n\n\n## Linting\n\nUse following command to make sure that your code is formatted per PEP8 spec:\n\n    poetry run task lint\n",
    'author': 'Eugen Ciur',
    'author_email': 'eugen@papermerge.com',
    'maintainer': 'Eugen Ciur',
    'maintainer_email': 'eugen@papermerge.com',
    'url': 'https://www.papermerge.com/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



import logging

from kombu.exceptions import OperationalError
from channels.layers import get_channel_layer
from pathlib import Path
from asgiref.sync import async_to_sync

from django.db.models.signals import post_save, post_delete, pre_delete
from celery.signals import (
    task_received,
    task_postrun,
    task_prerun,
    heartbeat_sent,
    worker_ready,
    worker_shutdown
)

from django.dispatch import receiver
from django.conf import settings
from papermerge.core.models import (
    Document,
    DocumentVersion,
    User,
)
from papermerge.core.storage import get_storage_instance
from .tasks import delete_user_data as delete_user_data_task
from .tasks import (
    ocr_document_task,
    post_ocr_document_task,
    generate_page_previews_task
)
from .signal_definitions import document_post_upload


logger = logging.getLogger(__name__)

# Tasks that need to notify websocket clients
MONITORED_TASKS = (
    'papermerge.core.tasks.ocr_document_task',
)

HEARTBEAT_FILE = Path("/tmp/worker_heartbeat")
READINESS_FILE = Path("/tmp/worker_ready")


@receiver(pre_delete, sender=Document)
def delete_files(sender, instance: Document, **kwargs):
    """
    Deletes physical (e.g. pdf) file associated
    with given (Document) instance.

    More exactly it will delete whatever it is inside
    associated folder in which original file was saved
    (e.g. all preview images).
    """
    for document_version in instance.versions.all():
        try:
            get_storage_instance().delete_doc(
                document_version.document_path
            )
        except IOError as error:
            logger.error(
                f"Error deleting associated file for document.pk={instance.pk}"
                f" {error}"
            )


@receiver(post_delete, sender=User)
def delete_user_data(sender, instance, **kwargs):
    """Deletes associated user folder(s) under media root"""
    delete_user_data_task.delay(str(instance.pk))


@receiver(post_save, sender=User)
def user_init(sender, instance, created, **kwargs):
    """
    Signal sent when user model is saved
    (create=True if user was actually created).
    Create user specific data when user is initially created
    """
    if created:
        if settings.PAPERMERGE_CREATE_SPECIAL_FOLDERS:
            instance.create_special_folders()


def get_channel_data(task_name, type):

    if task_name == 'papermerge.core.tasks.ocr_document_task':
        return {
            'type': f"ocrdocumenttask.{type}"
        }
    else:
        raise ValueError(f"Task name not in {MONITORED_TASKS}")


def channel_group_notify(task_name, task_kwargs, type):
    """
    Send group notification to the channel
    """
    channel_layer = get_channel_layer()
    channel_data = get_channel_data(task_name, type)

    channel_data.update(task_kwargs)
    task_short_name = task_name.split('.')[-1]

    logger.debug(
        f"channel_group_notify {task_short_name} {channel_data}"
    )
    async_to_sync(
        channel_layer.group_send
    )(
        task_short_name, channel_data
    )


@task_prerun.connect
def channel_group_notify_task_prerun(sender=None, **kwargs):
    if sender:
        if sender.name in MONITORED_TASKS:
            channel_group_notify(
                task_name=sender.name,
                task_kwargs=kwargs['kwargs'],
                type='taskstarted'
            )


@task_received.connect
def channel_group_notify_task_received(sender=None, **kwargs):
    request = kwargs.get('request')
    if request:
        if request.name in MONITORED_TASKS:
            channel_group_notify(
                task_name=request.name,
                task_kwargs=request.kwargs,
                type='taskreceived'
            )


@task_postrun.connect
def channel_group_notify_task_postrun(sender=None, **kwargs):
    if sender:
        if sender.name in MONITORED_TASKS:
            state = kwargs['state']
            if state == 'SUCCESS':
                type = 'tasksucceeded'
            else:
                type = 'taskfailed'

            channel_group_notify(
                task_name=sender.name,
                task_kwargs=kwargs['kwargs'],
                type=type
            )


@heartbeat_sent.connect
def heartbeat(**_):
    # liveness probe for celery worker
    # https://github.com/celery/celery/issues/4079
    HEARTBEAT_FILE.touch()


@worker_ready.connect
def worker_ready(**_):
    # readyness probe for celery worker
    # https://github.com/celery/celery/issues/4079
    READINESS_FILE.touch()


@worker_shutdown.connect
def worker_shutdown(**_):
    # https://github.com/celery/celery/issues/4079
    for file in (HEARTBEAT_FILE, READINESS_FILE):
        if file.is_file():
            file.unlink()


@receiver(document_post_upload, sender=Document)
def generate_page_previews(
    sender,
    document_version: DocumentVersion,
    **_
):
    """Generates page previews"""
    generate_page_previews_task.delay(str(document_version.id))


@receiver(document_post_upload, sender=Document)
def receiver_document_post_upload(
    sender,
    document_version: DocumentVersion,
    **_
):
    """
    Triggers document OCR if automatic OCR is on.

    Arguments:
        document - instance of associated document model
        document_version - instance of newly created document version
    """
    doc_ver = document_version
    doc = document_version.document
    user = doc.user

    logger.debug(
        "document_post_upload"
        f" [doc.id={doc.id}]"
        f" [doc_version.number={doc_ver.number}]"
        f" [doc_version_id={doc_ver.id}]"
        f" [user.id={user.id}]"
    )

    user_settings = user.preferences
    namespace = getattr(get_storage_instance(), 'namespace', None)

    if user_settings['ocr__trigger'] == 'auto':
        try:
            ocr_document_task.apply_async(
                kwargs={
                    'document_id': str(doc.id),
                    'lang': doc.lang,
                    'namespace': namespace,
                    'user_id': str(user.id)
                },
                link=[
                    post_ocr_document_task.s(namespace),
                ]
            )
        except OperationalError:
            # If redis service is not available then:
            # - request is accepted
            # - document is uploaded
            # - warning is logged
            # - response includes exception message text
            logger.warning(
                "Operation Error while creating the task",
                exc_info=True
            )



import logging

from asgiref.sync import async_to_sync
from channels.generic.websocket import JsonWebsocketConsumer

from papermerge.notifications.mixins import RequireAuth

logger = logging.getLogger(__name__)


class InboxRefreshConsumer(RequireAuth, JsonWebsocketConsumer):
    group_name = "inbox_refresh"

    def disconnect(self, close_code):
        async_to_sync(
            self.channel_layer.group_discard
        )(self.group_name, self.channel_name)

    def inbox_refresh(self, data):
        logger.debug("inbox_refresh event triggered")

        if data['user_id'] != str(self.user.id):
            # notification is intended only for users who
            # initiated document updates
            return

        logger.debug(
            f"inbox refresh data={data} for user_id={self.user.id}"
        )
        self.send_json(data)



import lxml.html


def parse_bbox_title(bbox_title):
    """Gets a string like
    input: 'bbox 23 344 45 66; x_wconf 88'
    output a hash:  {x1: 23, y1: 344, x2: 45, y2: 66, wconf: 88}

    Look here: http://kba.cloud/hocr-spec/1.2/
    for bbox and x_wconf properties.
    """
    pass


def extract_words_from(hocr_file):
    html = None
    result = []

    with open(hocr_file, "rb") as f:
        text = f.read()
        html = lxml.html.fromstring(text)

    for span in html.xpath("//span[@class='ocrx_word']"):
        elem = {}
        elem['class'] = span.attrib['class']
        elem['id'] = span.attrib['id']
        bbox = parse_bbox_title(span.attrib['title'])
        elem['title'] = span.attrib['title']
        elem['text'] = span.text
        elem['x1'] = bbox['x1']
        elem['x2'] = bbox['x1']
        elem['y1'] = bbox['y1']
        elem['y2'] = bbox['y2']
        elem['wconf'] = bbox['wconf']

        result.append(elem)

    return result



from django.db.models import QuerySet

from rest_framework.generics import GenericAPIView
from rest_framework.response import Response
from rest_framework.renderers import JSONRenderer

from drf_spectacular.utils import (
    extend_schema,
    OpenApiParameter
)
from haystack.query import SearchQuerySet, SQ
from papermerge.core.views.mixins import RequireAuthMixin
from papermerge.search.serializers import SearchResultSerializer
from papermerge.search.constants import (
    TAGS_OP_ALL,
    TAGS_OP_ANY
)


class SearchView(RequireAuthMixin, GenericAPIView):
    """
    Performs full text search on the documents and folders.

    Folders are matched by their title and assigned tags.
    Documents are matched by title, OCRed text and assigned tags.
    """
    resource_name = 'search'
    serializer_class = SearchResultSerializer
    renderer_classes = [JSONRenderer]

    @extend_schema(
        operation_id="Search",
        parameters=[
            OpenApiParameter(
                name='q',
                description='text to search',
                required=False,
                type=str,
            ),
            OpenApiParameter(
                name='tags',
                description=f"""
                Comma delimited tags that should be assigned the node.
                By default uses `{TAGS_OP_ALL}` operator i.e. all tags listed
                here should be assgned to the node. For `{TAGS_OP_ANY}` operator
                use `tags_ap={TAGS_OP_ANY}`
                """,
                required=False,
                type=str,
            ),
            OpenApiParameter(
                name='tags_op',
                description=f"""
                Operator to use when searching by tag.
                For `{TAGS_OP_ANY}` - will return nodes with at least one of
                    the tags assigned.
                For `{TAGS_OP_ALL}` - will return only nodes with all of the
                tags assigned.
                """,
                required=False,
                type=str,
                enum=[TAGS_OP_ALL, TAGS_OP_ANY],
                default=TAGS_OP_ALL
            )
        ],
        responses={
            '200': {
                'type': 'array',
                'items': {
                    'type': 'object',
                    'properties': {
                        'id': {
                            'type': 'string',
                            'format': 'uuid'
                        },
                        'text': {
                            'type': 'string',
                            'nullable': True
                        },
                        'title': {
                            'type': 'string'
                        },
                        'tags': {
                            'type': 'array',
                            'items': {
                                'type': 'string'
                            }
                        },
                        'breadcrumb': {
                            'type': 'array',
                            'items': {
                                'type': 'string'
                            }
                        },
                        'node_type': {
                            'type': 'string'
                        },
                        'highlight': {
                            'type': 'string',
                            'nullable': True
                        },
                        'user_id': {
                            'type': 'string',
                            'nullable': True,
                            'format': 'uuid'
                        }
                    }
                }
            }
        }
    )
    def get(self, request):
        query_text = request.query_params.get('q', '')
        if len(query_text) == 0:
            query_text = '*'
        query_tags = request.query_params.get('tags', '')
        tags_op = request.query_params.get('tags_op', TAGS_OP_ALL)
        #  never trust user input + make sure only valid options are used
        if tags_op not in (TAGS_OP_ALL, TAGS_OP_ANY):
            tags_op = TAGS_OP_ALL

        query_all = SearchQuerySet().filter(user=request.user)

        query_all = self.add_filter_by_tags(
            query=query_all,
            query_tags=query_tags,
            tags_op=tags_op
        )

        if query_text != '*':
            query_all = self.add_filter_by_content(
                query=query_all,
                query_text=query_text
            )

        query_all = query_all.highlight()
        serializer = SearchResultSerializer(query_all, many=True)

        return Response(serializer.data)

    def add_filter_by_content(
        self,
        query: SearchQuerySet,
        query_text: str
    ) -> SearchQuerySet:

        by_title = SQ(title__startswith=query_text.lower()) | SQ(
            title=query_text.lower()
        )
        by_content = SQ(last_version_text__contains=query_text) | SQ(
            last_version_text=query_text
        )

        return query.filter(by_content | by_title)

    def add_filter_by_tags(
        self,
        query: SearchQuerySet,
        query_tags: str,
        tags_op: str
    ) -> SearchQuerySet:

        if not query_tags:
            return query

        if tags_op == TAGS_OP_ALL:
            sq = SQ()
            for name in query_tags.split(','):
                sq = sq & SQ(tags__contain=name)
        else:
            # TAGS_OP_ANY
            sq = SQ()
            for name in query_tags.split(','):
                sq = sq | SQ(tags__contain=name)

        return query.filter(sq)

    def get_queryset(self):
        # This is workaround warning issued when runnnig
        # `./manage.py generateschema`
        # https://github.com/carltongibson/django-filter/issues/966
        if not self.request:
            return None

        queryset = self.queryset
        if isinstance(queryset, QuerySet):
            # Ensure queryset is re-evaluated on each request.
            queryset = queryset.all()
        return queryset



import io
import os
from typing import Optional, Union
from pikepdf import Pdf
from collections import abc, namedtuple

from django.utils.html import escape

from papermerge.core.lib.path import PagePath
from papermerge.core.storage import abs_path, get_storage_instance
from papermerge.core.models import DocumentVersion


def sanitize_kvstore(kvstore_dict):
    """
    Creates a sanitized dictionary.

    Sanitizied dictionary contains only allowed keys and escaped values.
    """
    allowed_keys = [
        'id',
        'key',
        'value',
        'kv_type',
        'kv_format',
        'kv_inherited',
    ]

    sanitized_kvstore_dict = {}

    for allowed_key in allowed_keys:
        if allowed_key in kvstore_dict.keys():
            value = kvstore_dict.get(allowed_key, None)
            if isinstance(value, bool):
                allowed_value = value
            else:
                allowed_value = escape(kvstore_dict.get(allowed_key, None))

            sanitized_kvstore_dict[allowed_key] = allowed_value

    return sanitized_kvstore_dict


def sanitize_kvstore_list(kvstore_list):
    """
    Creates a new list of sanitized dictionaries.

    Sanitizied dictionary contains only allowed keys and escaped values.
    """
    if not isinstance(kvstore_list, list):
        raise ValueError("Expects list type as input")

    new_kvstore_list = [
        sanitize_kvstore(item) for item in kvstore_list
    ]

    return new_kvstore_list


PageRecycleMapItem = namedtuple(
    'PageRecycleMapItem', ['new_number', 'old_number']
)


class PageRecycleMap:
    """Maps new page numbers to old ones.

    Under the hood, when user deletes pages from a document version, a new
    document version is created and only "not deleted" pages of old document
    version are transferred. This transfer means that data, like OCRed files,
    database fields, from "old page" are moved to the "new page". Thus,
    the requirement to hold a structure which maps page from old document
    version to the new document version.

    Couple of examples.
    Say there is 6 pages document and user deletes pages numbered 1 and 2;

    |Old document version | New document version|
    ---------------------------------------------
    |        1            |    3                |
    |        2            |    4                |
    |        3            |    5                |
    |        4            |    6                |
    |        5            |                     |
    |        6            |                     |
    ---------------------------------------------

    Notice that pages retain their original label. Table above reads
    as follows: in the new document version page number "1" gets
    info from old document version page number "3". New page number "2"
    gets info from old page number "4". New page number "3" gets info
    from page number "5". And finally the new page number "4" gets info from
    old page number "6".

    One more example; this time consider 5 pages document. User
    deletes page number 1 and 5:

    |Old document version | New document version|
    ---------------------------------------------
    |        1            |    2                |
    |        2            |    3                |
    |        3            |    4                |
    |        4            |                     |
    |        5            |                     |
    ---------------------------------------------

    The mapping will be:
        [
            {new_number: 1, old_number: 2},
            {new_number: 2, old_number: 3},
            {new_number: 3, old_number: 4},
        ]

    Last example: total pages = 5, deleted pages = [2, 3]

    |Old document version | New document version|
    ---------------------------------------------
    |        1            |    1                |
    |        2            |    4                |
    |        3            |    5                |
    |        4            |                     |
    |        5            |                     |
    ---------------------------------------------

    Result mapping is:
        [
            {new_number: 1, old_number: 1},
            {new_number: 2, old_number: 4},
            {new_number: 3, old_number: 5},
        ]
    """

    def __init__(self, total: int, deleted: list[int] = []):
        if not isinstance(deleted, abc.Sequence):
            raise ValueError('`deleted` expected to be a sequence')

        if total < len(deleted):
            raise ValueError('`total` < `deleted`')

        self.total = total
        self.deleted = deleted

        _pages = [
            page for page in range(1, self.total + 1)
            if page not in self.deleted
        ]
        _page_numbers = range(1, len(_pages) + 1)
        self.page_map = zip(_page_numbers, _pages)

    def __iter__(self):
        return self

    def __next__(self):
        item = next(self.page_map)
        if item:
            return PageRecycleMapItem(*item)

        raise StopIteration

    def __repr__(self):
        return (
            f"PageRecycleMap("
            f"total={self.total!r}, deleted={self.deleted!r}"
            f")"
        )


def collect_text_streams(
    version: DocumentVersion,
    page_numbers: list[int]
) -> list[io.StringIO]:
    """
    Returns list of texts of given page numbers from specified document version

    Each page's text is wrapped as io.StringIO instance.
    """
    pages_map = {page.number: page for page in version.pages.all()}

    result = [
        io.StringIO(pages_map[number].text)
        for number in page_numbers
    ]

    return result


def reuse_ocr_data_multi(
    src_old_version: DocumentVersion,
    dst_old_version: Optional[DocumentVersion],
    dst_new_version: DocumentVersion,
    page_numbers: list[int],
    position: int = 0
):
    if dst_old_version is None:
        position = 0

    storage = get_storage_instance()
    page_map = [(pos, pos) for pos in range(1, position + 1)]

    if len(page_map) > 0 and dst_old_version is not None:
        for src_page_number, dst_page_number in page_map:
            src_page_path = PagePath(
                document_path=dst_old_version.document_path,
                page_num=src_page_number
            )
            dst_page_path = PagePath(
                document_path=dst_new_version.document_path,
                page_num=dst_page_number
            )
        storage.copy_page(src=src_page_path, dst=dst_page_path)

    page_map = zip(
        page_numbers,
        [pos for pos in range(position + 1, position + len(page_numbers) + 1)]
    )

    for src_page_number, dst_page_number in page_map:
        src_page_path = PagePath(
            document_path=src_old_version.document_path,
            page_num=src_page_number
        )
        dst_page_path = PagePath(
            document_path=dst_new_version.document_path,
            page_num=dst_page_number
        )
        storage.copy_page(src=src_page_path, dst=dst_page_path)

    if dst_old_version is not None:
        dst_old_total_pages = dst_old_version.pages.count()
        _range = range(
            position + 1,
            dst_old_total_pages + 1
        )
        page_map = [(pos, pos + len(page_numbers)) for pos in _range]

        for src_page_number, dst_page_number in page_map:
            src_page_path = PagePath(
                document_path=dst_old_version.document_path,
                page_num=src_page_number
            )
            dst_page_path = PagePath(
                document_path=dst_new_version.document_path,
                page_num=dst_page_number
            )
            storage.copy_page(src=src_page_path, dst=dst_page_path)


def reuse_ocr_data(
    old_version: DocumentVersion,
    new_version: DocumentVersion,
    page_map: Union[PageRecycleMap, list]
) -> None:
    storage_instance = get_storage_instance()

    for new_number, old_number in page_map:
        src_page_path = PagePath(
            document_path=old_version.document_path,
            page_num=old_number
        )
        dst_page_path = PagePath(
            document_path=new_version.document_path,
            page_num=new_number
        )
        storage_instance.copy_page(
            src=src_page_path,
            dst=dst_page_path
        )


def reuse_text_field(
    old_version: DocumentVersion,
    new_version: DocumentVersion,
    page_map: list
) -> None:
    streams = collect_text_streams(
        version=old_version,
        # list of old_version page numbers
        page_numbers=[item[1] for item in page_map]
    )

    # updates page.text fields and document_version.text field
    new_version.update_text_field(streams)


def reuse_text_field_multi(
    src_old_version: DocumentVersion,
    dst_old_version: Optional[DocumentVersion],
    dst_new_version: DocumentVersion,
    page_numbers: list[int],
    position: int = 0,
):
    """
    Copies `text` field from two sources to the destination

    :param src_old_version: source 1
    :param dst_old_version: source 2
    :param dst_new_version: destination
    :param page_numbers: which pages from source 1 to copy
    :param position: at which position (in destination) to insert pages

    reuse_text_field_multi updates `text` field of destination document version
    and each of its associated pages.

    Note: page `position` starts with 0
    `page_numbers` is a list of page numbers. In this list, page numbering
    starts with 1.
    """
    if dst_old_version is None:
        position = 0

    page_map = [(pos, pos) for pos in range(1, position + 1)]
    streams = []
    if len(page_map) > 0 and dst_old_version is not None:
        streams.extend(
            collect_text_streams(
                version=dst_old_version,
                page_numbers=[item[1] for item in page_map]
            )
        )

    streams.extend(
        collect_text_streams(
            version=src_old_version,
            page_numbers=page_numbers
        )
    )

    if dst_old_version is not None:
        dst_new_total_pages = dst_new_version.pages.count()
        _range = range(
            position + 1 + len(page_numbers),
            dst_new_total_pages + 1
        )
        page_numbers_to_collect = [
            pos - len(page_numbers) for pos in _range
        ]
        streams.extend(
           collect_text_streams(
                version=dst_old_version,
                page_numbers=list(page_numbers_to_collect)
           )
        )

    dst_new_version.update_text_field(streams)


def remove_pdf_pages(
    old_version: DocumentVersion,
    new_version: DocumentVersion,
    page_numbers: list[int]
):
    """
    :param old_version: is instance of DocumentVersion
    :param new_version:  is instance of DocumentVersion
    :param page_numbers: numbers of pages to delete. Numbering starts with 1.

    Notice that page numbering starts with 1 i.e. page_numbers=[1, 2] -
    will remove first and second pages.
    """
    # delete page from document's new version associated file

    if len(page_numbers) < 1:
        raise ValueError("Empty page_numbers")

    pdf = Pdf.open(
        abs_path(old_version.document_path.url)
    )

    if len(pdf.pages) < len(page_numbers):
        raise ValueError("Too many values in page_numbers")

    _deleted_count = 0
    for page_number in page_numbers:
        pdf.pages.remove(p=page_number - _deleted_count)
        _deleted_count += 1

    dirname = os.path.dirname(
        abs_path(new_version.document_path.url)
    )
    os.makedirs(dirname, exist_ok=True)
    pdf.save(abs_path(new_version.document_path.url))


def insert_pdf_pages(
    src_old_version: DocumentVersion,
    dst_old_version: Optional[DocumentVersion],
    dst_new_version: DocumentVersion,
    src_page_numbers: list[int],
    dst_position: int = 0
) -> None:
    """Inserts pages from source to destination at given position

    In case both `dst_old_version` and `dst_new_version` parameters
    are non-empty `DocumentVersion` instances - `insert_pdf_pages` will take
    `src_page_numbers` from `src_old_version` and
    insert them at `dst_position` of `dst_old_version` and will
    save result in `dst_new_version`.

    In case `dst_old_version` is None - `insert_pdf_pages` will
    take `src_page_numbers` from `src_old_version` and insert
    at position 0 of the newly created pdf. Newly created pdf will be saved
    at `dst_new_version`.

    Remarks:
    `dst_position` starts with 0.
    In `src_page_numbers` page numbering starts with 1 i.e.
    when `src_page_numbers=[1, 2]` means insert first and second pages from
    source document version.
    """
    src_old_pdf = Pdf.open(
        abs_path(src_old_version.document_path.url)
    )
    if dst_old_version is None:
        # case of total merge
        dst_old_pdf = Pdf.new()
        dst_position = 0
    else:
        dst_old_pdf = Pdf.open(
            abs_path(dst_old_version.document_path.url)
        )

    _inserted_count = 0
    for page_number in src_page_numbers:
        pdf_page = src_old_pdf.pages.p(page_number)
        dst_old_pdf.pages.insert(dst_position + _inserted_count, pdf_page)
        _inserted_count += 1

    dirname = os.path.dirname(
        abs_path(dst_new_version.document_path.url)
    )
    os.makedirs(dirname, exist_ok=True)
    dst_old_pdf.save(
        abs_path(dst_new_version.document_path.url)
    )


def total_merge(
    src_old_version: DocumentVersion,
    dst_new_version: DocumentVersion
) -> None:
    """
    Merge source document version with destination

    'Total' means 'all pages'.
    """
    # all pages of the source
    page_numbers = [page.number for page in src_old_version.pages.all()]

    insert_pdf_pages(
        src_old_version=src_old_version,
        dst_old_version=None,
        dst_new_version=dst_new_version,
        src_page_numbers=page_numbers,
        dst_position=0
    )
    reuse_ocr_data_multi(
        src_old_version=src_old_version,
        dst_old_version=None,
        dst_new_version=dst_new_version,
        position=0,
        page_numbers=page_numbers
    )

    reuse_text_field_multi(
        src_old_version=src_old_version,
        dst_old_version=None,
        dst_new_version=dst_new_version,
        position=0,
        page_numbers=page_numbers
    )
    # Total merge deletes source document.
    # Because all pages of the source are moved to destination, source's
    # last version remains "without pages". A document version without pages
    # does not make sense to stay around - thus we delete it!
    src_old_version.document.delete()


def partial_merge(
    src_old_version: DocumentVersion,
    src_new_version: DocumentVersion,
    dst_new_version: DocumentVersion,
    page_numbers: list[int]
) -> None:
    """Merge only some pages of the source document version with destination

    No all pages of the source are used, which means
    source document version IS NOT DELETED.

    'Partial' means 'not all pages'.
    """

    if len(page_numbers) >= src_old_version.pages.count():
        raise ValueError("Number of pages to remove exceeds source page count")

    # remove pages from the source document version
    remove_pdf_pages(
        old_version=src_old_version,
        new_version=src_new_version,
        page_numbers=page_numbers
    )

    page_map = list(
        PageRecycleMap(
            total=src_old_version.page_count,
            deleted=page_numbers
        )
    )

    reuse_ocr_data(
        old_version=src_old_version,
        new_version=src_new_version,
        page_map=page_map
    )

    reuse_text_field(
        old_version=src_old_version,
        new_version=src_new_version,
        page_map=page_map
    )

    # insert pages to the destination
    insert_pdf_pages(
        src_old_version=src_old_version,
        dst_old_version=None,
        dst_new_version=dst_new_version,
        src_page_numbers=page_numbers
    )

    reuse_ocr_data_multi(
        src_old_version=src_old_version,
        dst_old_version=None,
        dst_new_version=dst_new_version,
        position=0,
        page_numbers=page_numbers
    )

    reuse_text_field_multi(
        src_old_version=src_old_version,
        dst_old_version=None,
        dst_new_version=dst_new_version,
        position=0,
        page_numbers=page_numbers
    )


def reorder_pdf_pages(
    old_version,
    new_version,
    pages_data,
    page_count
):
    src = Pdf.open(abs_path(old_version.document_path.url))

    dst = Pdf.new()
    reodered_list = sorted(pages_data, key=lambda item: item['new_number'])

    for list_item in reodered_list:
        page = src.pages.p(list_item['old_number'])
        dst.pages.append(page)

    dirname = os.path.dirname(
        abs_path(new_version.document_path.url)
    )
    os.makedirs(dirname, exist_ok=True)
    dst.save(abs_path(new_version.document_path.url))


def rotate_pdf_pages(
    old_version,
    new_version,
    pages_data
):
    """
    ``pages`` data is a list of dictionaries. Each dictionary is expected
    to have following keys:
        - number
        - angle
    """
    src = Pdf.open(abs_path(old_version.document_path.url))

    for page_data in pages_data:
        page = src.pages.p(page_data['number'])
        page.rotate(page_data['angle'], relative=True)

    dirname = os.path.dirname(
        abs_path(new_version.document_path.url)
    )
    os.makedirs(dirname, exist_ok=True)
    src.save(abs_path(new_version.document_path.url))



from django.contrib.auth.models import ContentType
from rest_framework import generics
from drf_spectacular.utils import extend_schema

from papermerge.core.serializers import ContentTypeSerializer


@extend_schema(exclude=True)
class ContentTypeRetrieve(generics.RetrieveAPIView):
    serializer_class = ContentTypeSerializer
    pagination_class = None
    queryset = ContentType.objects.all()



from rest_framework_json_api.pagination import JsonApiPageNumberPagination


class JsonApiPagination(JsonApiPageNumberPagination):
    """
    A JSON:API compatible pagination with schema included
    """

    def get_paginated_response_schema(self, schema):
        return {
            'type': 'object',
            'properties': {
                'data': schema,
                'meta': {
                    'type': 'object',
                    'properties': {
                        'pagination': {
                            'type': 'object',
                            'properties': {
                                'page': {
                                    'type': 'integer',
                                    'example': 1
                                },
                                'pages': {
                                    'type': 'integer',
                                    'example': 10
                                },
                                'count': {
                                    'type': 'integer',
                                    'example': 100
                                }
                            }
                        }
                    }
                },
                'links': {
                    'type': 'object',
                    'properties': {
                        'next': {
                            'type': 'string',
                            'nullable': True,
                            'format': 'uri',
                            'example': 'http://api.example.org/accounts/?{page_query_param}=4'.format(  # noqa
                                page_query_param=self.page_query_param)
                        },
                        'prev': {
                            'type': 'string',
                            'nullable': True,
                            'format': 'uri',
                            'example': 'http://api.example.org/accounts/?{page_query_param}=2'.format(   # noqa
                                page_query_param=self.page_query_param)
                        },
                        'first': {
                            'type': 'string',
                            'nullable': True,
                            'format': 'uri',
                            'example': 'http://api.example.org/accounts/?{page_query_param}=2'.format(   # noqa
                                page_query_param=self.page_query_param)
                        },
                        'last': {
                            'type': 'string',
                            'nullable': True,
                            'format': 'uri',
                            'example': 'http://api.example.org/accounts/?{page_query_param}=2'.format(  # noqa
                                page_query_param=self.page_query_param)
                        },
                    }
                },
            }
        }



# Generated by Django 4.0.8 on 2022-12-30 07:04

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0005_basetreenode_unique title per parent'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='basetreenode',
            options={'verbose_name': 'Documents', 'verbose_name_plural': 'Documents'},
        ),
        migrations.RemoveField(
            model_name='basetreenode',
            name='level',
        ),
        migrations.RemoveField(
            model_name='basetreenode',
            name='lft',
        ),
        migrations.RemoveField(
            model_name='basetreenode',
            name='polymorphic_ctype',
        ),
        migrations.RemoveField(
            model_name='basetreenode',
            name='rght',
        ),
        migrations.RemoveField(
            model_name='basetreenode',
            name='tree_id',
        ),
        migrations.AddField(
            model_name='basetreenode',
            name='ctype',
            field=models.CharField(blank=True, choices=[('folder', 'folder'), ('document', 'document')], max_length=16, null=True),
        ),
    ]




from django.contrib.auth import get_user_model
from rest_framework.permissions import DjangoModelPermissions

# custom user is used - papermerge.core.models.User
User = get_user_model()


class CustomModelPermissions(DjangoModelPermissions):
    """
    The request is authenticated using `django.contrib.auth` permissions.
    See: https://docs.djangoproject.com/en/dev/topics/auth/#permissions

    It ensures that the user is authenticated, and has the appropriate
    `view`/`add`/`change`/`delete` permissions on the model.
    """
    # Overrides perms_map of `DjangoModelPermissions` with value for 'GET',
    # 'HEAD' and 'OPTIONS' keys
    perms_map = {
        'GET': ['%(app_label)s.view_%(model_name)s'],
        'OPTIONS': ['%(app_label)s.view_%(model_name)s'],
        'HEAD': ['%(app_label)s.view_%(model_name)s'],
        'POST': ['%(app_label)s.add_%(model_name)s'],
        'PUT': ['%(app_label)s.change_%(model_name)s'],
        'PATCH': ['%(app_label)s.change_%(model_name)s'],
        'DELETE': ['%(app_label)s.delete_%(model_name)s'],
    }



from django.contrib.auth.models import Group, Permission

from rest_framework.serializers import Serializer
from rest_framework_json_api.serializers import ModelSerializer

from .permission import PermissionSerializer


class GroupSerializer(ModelSerializer):

    included_serializers = {
        'permissions': PermissionSerializer,
    }

    class Meta:
        model = Group
        resource_name = 'groups'
        fields = (
            'id',
            'name',
            'permissions',
        )

    class JSONAPIMeta:
        included_resources = ['permissions']

    def create(self, validated_data):
        perm_ids = [perm.id for perm in validated_data.get('permissions', [])]
        permissions = Permission.objects.filter(id__in=perm_ids)
        group = Group.objects.create(
            name=validated_data['name']
        )
        group.permissions.set(permissions)
        group.save()

        return group

    def update(self, instance, validated_data):
        perm_ids = [perm.id for perm in validated_data['permissions']]
        permissions = Permission.objects.filter(id__in=perm_ids)
        instance.permissions.set(permissions)
        instance.name = validated_data['name']
        instance.save()

        return instance


class Data_GroupSerializer(Serializer):
    data = GroupSerializer()



from django.db import models

from papermerge.core.models.node import BaseTreeNode


class FolderManager(models.Manager):
    pass


class FolderQuerySet(models.QuerySet):

    def delete(self, *args, **kwargs):
        for node in self:
            descendants = node.get_descendants()

            if len(descendants) > 0:
                descendants.delete(*args, **kwargs)
            # At this point all descendants were deleted.
            # Self delete :)
            try:
                node.delete(*args, **kwargs)
            except BaseTreeNode.DoesNotExist:
                # this node was deleted by earlier recursive call
                # it is ok, just skip
                pass


CustomFolderManager = FolderManager.from_queryset(FolderQuerySet)


class Folder(BaseTreeNode):

    # Each user has two special folders: Inbox and Home. Each of this folders
    # is created when respective user model is created.
    # Inbox is where all incoming documents are placed (i.e. from IMAP client).
    # Home is topmost folder from his/her stand point i.e. 'landing page'
    # when he/she lands in web UI. Everything outside user's home and
    # inbox folders is not visible/accessible to him/her.
    # Both Inbox and Home are user's topmost folders i.e.
    # is_parent = null and user_id = <current_user_id>

    # Special folders' name always starts with a DOT ('.') character
    INBOX_TITLE = ".inbox"
    HOME_TITLE = ".home"

    objects = CustomFolderManager()

    class JSONAPIMeta:
        resource_name = "folders"

    def delete(self, *args, **kwargs):
        descendants = self.basetreenode_ptr.get_descendants()

        if len(descendants) > 0:
            for node in descendants:
                try:
                    node.delete(*args, **kwargs)
                except BaseTreeNode.DoesNotExist:
                    pass
        # At this point all descendants were deleted.
        # Self delete :)
        try:
            super().delete(*args, **kwargs)
        except BaseTreeNode.DoesNotExist:
            # this node was deleted by earlier recursive call
            # it is ok, just skip
            pass

    class Meta:
        verbose_name = "Folder"
        verbose_name_plural = "Folders"

    def __str__(self):
        return self.title

    def get_children(self):
        """Returns direct children of current node"""
        return Folder.objects.filter(parent=self)


def get_inbox_children(user):
    """
    Returns a ``QuerySet`` containing the immediate children nodes
    of given user's inbox folder
    """
    inbox_node = BaseTreeNode.objects.get(
        title=Folder.INBOX_TITLE,
        user=user
    )

    return inbox_node.get_children()



import logging

from django.core.management import BaseCommand
from papermerge.core.backup_restore import (
    _can_restore,
    restore_documents
)
from papermerge.core.models import User

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = """
        Restore all Documents and their folder structure from an archive.
        If you pass a username with --user it will restore 'per user backup'.
        If you don't pass username with --user/-u - it is assumend that you
        want to restore "all users" backup.

        By default will trigger OCR for each imported document.
        Use --skip-ocr if you wish to OCR documents later.
    """

    def add_arguments(self, parser):

        parser.add_argument(
            '--user', type=str,
            help="user (username of) the restored documents should belong to",
            default=None
        )

        parser.add_argument('location', nargs='?', type=str)

    def handle(self, *args, **options):

        username = options.get('user')

        user = None
        if username:
            try:
                user = User.objects.get(username=username)
            except User.DoesNotExist:
                logger.error(
                    f"Username {username} not found."
                )
                return

        if options.get('location') is None:
            logger.error("Please add the path to your backup.tar")
        else:
            with open(options.get('location'), 'rb') as restore_file:
                if _can_restore(restore_file):
                    restore_documents(
                        restore_file=restore_file,
                        user=user,
                        skip_ocr=True
                    )
                else:
                    logging.error(
                        "Archive cannot be restored because of"
                        " version mismatch."
                    )



from rest_framework import serializers as rest_serializers
from rest_framework_json_api import serializers

from django.urls import reverse
from drf_spectacular.utils import extend_schema_field
from drf_spectacular.openapi import OpenApiTypes

from papermerge.core.models import Page


class PageSerializer(serializers.ModelSerializer):

    svg_url = serializers.SerializerMethodField()
    jpg_url = serializers.SerializerMethodField()

    class Meta:
        model = Page
        resource_name = 'pages'
        fields = (
            'id',
            'number',
            'text',
            'lang',
            'document_version',
            'svg_url',
            'jpg_url'
        )

    @extend_schema_field(OpenApiTypes.STR)
    def get_svg_url(self, obj):
        return reverse('pages_page', args=[str(obj.pk)])

    @extend_schema_field(OpenApiTypes.STR)
    def get_jpg_url(self, obj):
        return reverse('pages_page', args=[str(obj.pk)])


class PageDeleteSerializer(rest_serializers.Serializer):
    # list of pages to delete
    pages = rest_serializers.ListField(
        child=rest_serializers.UUIDField()
    )


class PageReorderSerializer(rest_serializers.Serializer):
    id = rest_serializers.UUIDField()
    old_number = rest_serializers.IntegerField(
        help_text='Page position within the document before '
        " page's order change."
        'Position numbering starts with 1.'
    )
    new_number = rest_serializers.IntegerField(
        help_text='Desired new page position within the document. '
        'Position numbering starts with 1.'
    )


class PagesReorderSerializer(rest_serializers.Serializer):
    pages = PageReorderSerializer(many=True)


class PageRotateSerializer(rest_serializers.Serializer):
    id = rest_serializers.UUIDField()
    # rotation angle
    angle = rest_serializers.IntegerField()


class PagesRotateSerializer(rest_serializers.Serializer):
    pages = PageRotateSerializer(many=True)


class PagesMoveToFolderSerializer(rest_serializers.Serializer):
    pages = serializers.ListSerializer(
        child=serializers.UUIDField()
    )
    # destination folder node
    dst = rest_serializers.UUIDField()
    single_page = rest_serializers.BooleanField(default=False)
    title_format = rest_serializers.CharField(
        max_length=32,
        required=False
    )


class PagesMoveToDocumentSerializer(rest_serializers.Serializer):
    pages = serializers.ListSerializer(
        child=serializers.UUIDField()
    )
    # destination document node
    dst = rest_serializers.UUIDField()
    # If`merge` is True - ONLY new pages will be inserted in the
    # newly created document version.
    merge = rest_serializers.BooleanField(default=False)
    # if `merge` is True, then `position` field is discarded
    # if `merge` is False, then `position` field indicates
    # at which position should new pages be inserted:
    # 0 - at the beginning
    # 1 - after first page
    # 2 - after second page
    # ...
    # -1 - at the end
    position = rest_serializers.IntegerField(default=-1)
